import pandas as pd
import numpy as np
import os.path as op
from datetime import datetime, date

from .alerts_trader import find_last_trade, option_date
from .configurator import cfg

def get_date():
    time_strf = "%Y-%m-%d %H:%M:%S.%f"
    date = datetime.now().strftime(time_strf)
    return date


class AlertsTracker():

    def __init__(self, brokerage=None,
                 portfolio_fname=cfg['portfolio_names']["tracker_portfolio_name"],
                 dir_quotes = cfg['general']['data_dir'] + '/live_quotes' ):

        self.portfolio_fname = portfolio_fname  
        self.dir_quotes = dir_quotes
        self.bksession = brokerage    

        if op.exists(self.portfolio_fname):
            self.portfolio = pd.read_csv(self.portfolio_fname)
        else:
            self.portfolio = pd.DataFrame(columns=cfg["col_names"]['tracker_portfolio'].split(",") )
            self.portfolio.to_csv(self.portfolio_fname, index=False)

    def price_now(self, symbol:str, price_type="BTO"):
        if self.bksession is None:
            return None

        if price_type in ["BTO", "BTC"]:
            ptype = 'askPrice'
        else:
            ptype = 'bidPrice'

        quote = self.bksession.get_quotes([symbol])
        if quote is None:
            # Try it again in case of TDA
            quote = self.bksession.get_quotes([symbol])
            
        if quote is not None and len(quote) and quote.get(symbol).get('description' ) != 'Symbol not found':
            quote = quote.get(symbol).get(ptype)
            if quote != 0:
                return quote
            else:
                print("quote 0 for", symbol, price_type)

    def trade_alert(self, order, live_alert=True, channel=None):
        open_trade, _ = find_last_trade(order, self.portfolio, open_only=True)
        if order.get('uQty') is None:
            order['uQty'] = 1
        
        if order["action"] in ["BTO", "STC"] and live_alert:
            if order.get('Actual Cost', 'None') == 'None':
                order["Actual Cost"] = self.price_now(order["Symbol"], order["action"])

        if open_trade is None and order["action"] == "BTO":
            str_act = self.make_BTO(order, channel)
        elif order["action"] == "BTO":
            # str_act = "BTO averaging disabled as it is mostly wrong alert messages"
            str_act = self.make_BTO_Avg(order, open_trade)
        elif order["action"] == "STC" and open_trade is None:
            str_act = "STC without BTO"
        elif order["action"] == "STC":
            str_act = self.make_STC(order, open_trade)
        elif order["action"] == "ExitUpdate":
            if open_trade is not None:
                self.portfolio.loc[open_trade, "SL"] = order.get('SL')
            return 

        #save to csv
        self.portfolio.to_csv(self.portfolio_fname, index=False)      
        return str_act

    def make_BTO(self, order, chan=None):
        if order["price"] is None:
            return
        date = order.get("Date", get_date())
        if order['uQty'] is None:
            order['uQty'] = 1
        new_trade = {
            "Date": date,
            "Symbol": order['Symbol'],
            'isOpen': 1,
            "Asset": order["asset"],
            "Type": "BTO",
            "Price": order["price"],
            "Amount": order['uQty'],
            "Price-current": order.get("Actual Cost"),
            "Trader": order['Trader'],
            "SL": order.get("SL"),
            "Channel" : chan
            }        
        self.portfolio =pd.concat([self.portfolio, pd.DataFrame.from_records(new_trade, index=[0])], ignore_index=True)

        str_act = f"BTO {order['Symbol']} {order['price']}"
        if order['SL'] is not None:
            str_act += f", SL:{order['SL']}"
        return str_act

    def make_BTO_Avg(self, order, open_trade):
        current_Avg = self.portfolio.loc[open_trade, "Avged"]
        if np.isnan(current_Avg):
            current_Avg = 1
        else:
            current_Avg = int(current_Avg + 1)
        
        old_price = self.portfolio.loc[open_trade, "Price"]
        old_price = eval(old_price) if isinstance(old_price, str) else old_price
        old_qty = self.portfolio.loc[open_trade, "Amount"]
        alert_price_old = self.portfolio.loc[open_trade, "Price-current"]
        alert_price_old = eval(alert_price_old) if isinstance(alert_price_old, str) else alert_price_old
        alert_price_old = None if pd.isnull(alert_price_old) else alert_price_old
        alert_price = order.get("Actual Cost", "None")
        avgs_prices_al = f"{alert_price_old}/{alert_price}".replace("None/", "").replace("/None", "").replace("None", "")
        if not len(avgs_prices_al):
            avgs_prices_al = None
    
        self.portfolio.loc[open_trade, "Avged"] = current_Avg
        self.portfolio.loc[open_trade, "Amount"] += order['uQty']
        self.portfolio.loc[open_trade, "Price"] = round(((old_price*old_qty) + (order['price']*order['uQty']))/(old_qty+order['uQty']),2)
        self.portfolio.loc[open_trade, "Prices"] = f"{old_price}/{order['price']}"
        if alert_price == 'None' or alert_price is None or alert_price_old is None:
            self.portfolio.loc[open_trade, "Price-current"] = alert_price_old
        else:
            self.portfolio.loc[open_trade, "Price-current"] = round(((alert_price_old*old_qty) + (alert_price*order['uQty']))/(old_qty+order['uQty']),2)
        self.portfolio.loc[open_trade, "Prices-current"] = avgs_prices_al
        if order.get("SL"):
            self.portfolio.loc[open_trade, "SL"] =  order.get("SL")
        
        str_act = f"BTO {order['Symbol']} {current_Avg}th averging down @ {order['price']}"
        return str_act

    def make_STC(self, order, open_trade, check_trail=False):
        trade = self.portfolio.loc[open_trade]
        stc_info = calc_stc_prices(trade, order)
        #Log portfolio
        for k, v in stc_info.items():
            self.portfolio.loc[open_trade, k] = v
        
        trailstat = self.compute_trail(open_trade)
        self.portfolio.loc[open_trade, "TrailStats"] = trailstat
        self.portfolio.loc[open_trade, "STC-Date"] = order["Date"]
        stc_price =  self.portfolio.loc[open_trade,"STC-Price"]
        stc_utotal = self.portfolio.loc[open_trade,"STC-Amount"]
        suffx = ''
        if stc_utotal >= trade['Amount']:
            suffx = " Closed"
            self.portfolio.loc[open_trade, "isOpen"] = 0
            
        if stc_price == "none" or stc_price is None:
            str_STC = f"STC {order['Symbol']}  ({order['uQty']}), no price provided" + suffx
        else:
            str_STC = f"STC {order['Symbol']} ({order['uQty']}),{suffx} {stc_price:.2f}"
            if stc_info['STC-Price-current'] is not None:           
                       str_STC += f" current: {stc_info['STC-Price-current']:.2f} " 
            if stc_info["STC-PnL"] is not None:
                str_STC += f'PnL:{round(stc_info["STC-PnL"])}% ${round(stc_info["STC-PnL$"])}' 
            if stc_info["STC-PnL-current"] is not None:
                str_STC += f' Actual:{round(stc_info["STC-PnL-current"])}% ${round(stc_info["STC-PnL$-current"])}\n\t\t'

        if eval(order.get('# Closed', "0"))==1 :
            self.portfolio.loc[open_trade, "isOpen"]=0
        str_STC = str_STC + " " + trailstat.replace('|', '\n\t\t')
        return str_STC

    def compute_trail(self, open_trade):
        trade = self.portfolio.loc[open_trade]
        fname = self.dir_quotes + f"/{trade['Symbol']}.csv"
        if not op.exists(fname):
            return ""
        
        quotes = pd.read_csv(fname, on_bad_lines='skip')
        # start after BTO date
        dates = quotes['timestamp'].apply(lambda x: datetime.fromtimestamp(x))
        msk = dates >= pd.to_datetime(trade['Date'])
        quotes = quotes[msk].reset_index(drop=True)
        
        # first price will be the actual
        if not pd.isnull(trade["Price-current"]):
            price0 = trade["Price-current"]
        else:
            price0 = trade["Price"]
        quotes.loc[0, ' quote'] = price0

        # Calculate trailingStops
        res_str = "TS:{},{}%,${},in {}"
        trailing_Stop = [.2, .3, .4, .5] 
        max_trails = []      
        quotes['highest'] = quotes[' quote'].cummax() #take the cumulative max
        quotes['perc'] = round((quotes[' quote'] - quotes[' quote'].loc[0])/ quotes[' quote'].loc[0] *100,2)
        for trl in trailing_Stop:
            quotes[f'trailingstop{trl}'] = quotes['highest']*(1-trl) 
            trl_ix = (quotes[f' quote'] <= quotes[f'trailingstop{trl}']).idxmax()
            if trl_ix:
                tdiff_str, trl_r = self.trailing_get_time(trade['Date'], quotes, trl_ix)
                max_trails.extend([res_str.format(trl,trl_r['perc'],trl_r[' quote'],tdiff_str)])
        max_trails = "| ".join(max_trails)
        # get min max and their time
        quotes_stats = ""
        for st, ix in  zip(['min', 'max'],[ quotes['perc'].idxmin(),  quotes['perc'].idxmax()]):
            tdiff_str, trl_r = self.trailing_get_time(trade['Date'], quotes, ix)
            quotes_stats += f"{st},{trl_r['perc']}%,${trl_r[' quote']},in {tdiff_str}| "
        quotes_stats +=  "| " + max_trails
        return quotes_stats

    def trailing_get_time(self, trade_date, quotes, inx):
        trl_r = quotes.loc[inx]
        if pd.isna(trl_r['timestamp']):
            return "", trl_r
        tdiff =  datetime.fromtimestamp(trl_r['timestamp']) - pd.to_datetime(trade_date)
        tdiff_str = str(tdiff.round('s')).replace('0 days ','')
        return tdiff_str, trl_r

    def close_expired(self):
        for i, trade in  self.portfolio.iterrows():
            if trade["Asset"] != "option" or trade["isOpen"] == 0:
                continue
            optdate = option_date(trade['Symbol'])
            if optdate.date() < date.today():
                expdate = date.today().strftime("%Y-%m-%dT%H:%M:%S+0000")
                
                stc_info = calc_stc_prices(trade)
                #Log portfolio
                self.portfolio.loc[i, "STC-Date"] = expdate
                self.portfolio.loc[i, "STC-xQty"] = 1
                for k, v in stc_info.items():
                    self.portfolio.loc[i, k] = v
                
                self.portfolio.loc[i, "isOpen"] = 0
                str_prt = f"{trade['Symbol']} option expired -100%"
                print(str_prt)
        self.portfolio.to_csv(self.portfolio_fname, index=False)



def calc_stc_prices(trade, order=None):
    # if order is None = expired option
    if order is None:
        order = {
            "price":0,
            "Actual Cost":0,
            "expired": True
            }
    bto_price = trade["Price"]
    bto_price_al = trade["Price-current"]
    
    if not order.get('expired', False):
        if order['uQty'] is None:
            uQty = 1
        else:
            uQty = order['uQty']
    else: 
        if pd.isnull(trade["STC-Amount"]):
            uQty = trade['Amount'] 
        else:
            uQty = trade['Amount'] - trade["STC-Amount"]
    
    if isinstance(bto_price, str):
        bto_price =  np.mean(eval(bto_price.replace("/", ",")))
    if isinstance(bto_price_al, str):
        if bto_price_al[0] == '/':
            bto_price_al = bto_price_al[1:]
        bto_price_al =  np.mean(eval(bto_price_al.replace("/", ",")))
    
    if not pd.isnull(trade["STC-Price"]):  # previous stcs            
        stc_wprice = trade["STC-Price"] * trade["STC-Amount"]
        stc_utotal = trade["STC-Amount"] + uQty
        stc_price = (order.get("price") * uQty +  stc_wprice)/stc_utotal      
        prices = "/".join([str(trade["STC-Prices"]), str(order.get("price"))])
        
        if pd.isnull(trade["STC-Price-current"]) or pd.isnull(order.get("Actual Cost")):
            prices_curr = 0 if order.get('expired', False) else ""
            stc_price_al = 0 if order.get('expired', False) else None
        else:
            stc_wprice = trade["STC-Price-current"] * trade["STC-Amount"]
            stc_price_al = (order.get("Actual Cost") * uQty +  stc_wprice)/stc_utotal
            prices_curr = "/".join([str(trade["STC-Prices-current"]), str(order.get("Actual Cost"))])
    else:  # non-previous stcs   
        stc_price = order.get("price")
        stc_price_al = order.get("Actual Cost")
        stc_utotal = uQty
        prices = order.get("price")
        prices_curr = order.get("Actual Cost")
    
    mutipl = 1 if trade['Asset'] == "option" else .01  # pnl already in %
    if stc_price is not None: 
        stc_pnl = float((stc_price - bto_price)/bto_price) *100
        stc_pnl_u = stc_pnl* bto_price *mutipl*stc_utotal 
    else:
        stc_pnl = None
        stc_pnl_u = None
    
    if stc_price_al is None or pd.isnull(bto_price_al) or bto_price_al == 0:
        stc_pnl_al = None
        stc_pnl_al_u = None
    else:
        stc_pnl_al = float((stc_price_al - bto_price_al)/bto_price_al) *100
        stc_pnl_al_u = stc_pnl_al* bto_price_al *mutipl*stc_utotal 

    stc_info = {"STC-Prices":prices,
                "STC-Prices-current": prices_curr,
                "STC-Price": stc_price,
                "STC-Price-current": stc_price_al,
                "STC-Amount": stc_utotal,
                "STC-PnL": stc_pnl,
                "STC-PnL-current": stc_pnl_al,
                "STC-PnL$": stc_pnl_u,
                "STC-PnL$-current": stc_pnl_al_u,
                }
    return stc_info



import unittest
from unittest.mock import MagicMock
import pandas as pd
import os
from datetime import datetime, timedelta
from DiscordAlertsTrader.discord_bot import DiscordBot


class TestDiscordBot(unittest.TestCase):
    def test_new_msg_acts(self):
        self.tracker_portfolio_fname="test_tracker_portfolio.csv"
        queue_prints = MagicMock()
        bot = DiscordBot(queue_prints=queue_prints, live_quotes=False, brokerage=None,
                         tracker_portfolio_fname=self.tracker_portfolio_fname)
        expdate = datetime.now().strftime("%m/%d")
        message = pd.Series({'AuthorID': None,
                            'Author': "JonP",
                            'Date': "2022-01-01 10:00:00.000000", 
                            'Content': f'BTO 5 AI 25c {expdate} @ 1 <@&940418825235619910> swinging',
                            'Channel': "channel 1"
                            })
        bot.new_msg_acts(message, from_disc=False)

        # Example assertions for queue_prints
        print("here:", queue_prints.put.call_args_list)
        self.assertEqual(queue_prints.put.call_count, 3)
        self.assertEqual(queue_prints.put.call_args_list[0][0][0],
                         [f'2022-01-01 10:00:00 \t JonP: BTO 5 AI 25c {expdate} @ 1 <@&940418825235619910> swinging ', 'blue'])
        self.assertEqual(queue_prints.put.call_args_list[1][0][0],
                         [f'\t \t BTO AI {expdate} 25C @1.0 amount: 5', 'green'])

        # Delete the generated file
        os.remove(self.tracker_portfolio_fname)

    def test_new_msg_acts_wrong_date(self):
        self.tracker_portfolio_fname="test_tracker_portfolio.csv"
        queue_prints = MagicMock()
        bot = DiscordBot(queue_prints=queue_prints, live_quotes=False, brokerage=None,
                         tracker_portfolio_fname=self.tracker_portfolio_fname)
        expdate = datetime.now() - timedelta(days=1)
        expdate = expdate.strftime("%m/%d")
        
        message = pd.Series({'AuthorID': None,
                            'Author': "JonP",
                            'Date': "2022-01-01 10:00:00.000000", 
                            'Content': f'BTO 5 AI 25c {expdate} @ 1 <@&940418825235619910> swinging',
                            'Channel': "channel 1"
                            })
        bot.new_msg_acts(message, from_disc=False)

        # Example assertions for queue_prints
        print("here:", queue_prints.put.call_args_list)
        self.assertEqual(queue_prints.put.call_count, 2)
        self.assertEqual(queue_prints.put.call_args_list[0][0][0],
                         [f'2022-01-01 10:00:00 \t JonP: BTO 5 AI 25c {expdate} @ 1 <@&940418825235619910> swinging ', 'blue'])
        self.assertEqual(queue_prints.put.call_args_list[1][0][0],
                         [f'\t \t Option date in the past: {expdate}', 'green'])

        # Delete the generated file
        os.remove(self.tracker_portfolio_fname)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Apr  9 09:53:44 2021

@author: adonay
"""

import os.path as op
import pandas as pd
from datetime import datetime
import numpy as np
from .configurator import cfg
from .alerts_tracker import calc_stc_prices

def short_date(datestr, infrm="%Y-%m-%d %H:%M:%S.%f", outfrm="%m/%d/%Y %H:%M"):
    return datetime.strptime(datestr, infrm).strftime(outfrm)

def formt_num_2str(x, decim=2, str_len=6, remove_zero=True):
    if pd.isnull(x) or x == 'nan':
        return ""
    if remove_zero and x == 0:
        return ""
    x = round(x, decim)
    return f'{x: >{str_len}.{decim}f}'#.replace(".00", "   ")

def max_dig_len(values, decim=2):
    # Gives interger and decimal lengths in an array-like values
    values_s = [str(round(v, decim)) if v %1 else str(round(v)) for v in values]
    tmp_int = [len(v.split('.')[0]) for v in values_s]
    len_int = max(tmp_int) if tmp_int else 1
    tmp_dig = [len(v.split('.')[-1]) if "." in v else 0 for v in values_s]
    len_dig = max(tmp_dig) if tmp_dig else 1
    len_tot = len_int + len_dig + 1 if len_dig else  len_int + len_dig
    return len_tot, len_int, len_dig

def pd_col_str_frmt(pd_Series, max_decim=2, remove_zero=True):
    slen,_, decim = max_dig_len(pd_Series.to_numpy(), max_decim)
    return pd_Series.apply(lambda x: formt_num_2str(x,decim, slen, remove_zero))

def dataframe_num2str(db):
    for col in db:
        if db[col].dtype in [float, int]:
            db[col] = pd_col_str_frmt(db[col])
    return db

def round_int_flt(x, n=1):
    if x%1 == 0:
        return round(x)
    else:
        return round(x, n)

def format_exitplan(plan):

    if plan in ["", "{}"] or plan is None:
        return "None"
    plan = eval(plan)

    PT = [str(v) for v in [plan.get(f"PT{i}") for i in range(1,4)] if v is not None]
    SL = plan.get('SL')

    plan = "PT:" + ",".join(PT) if PT else ""
    sl_str = ", SL:" if plan else "SL:"
    plan = plan + sl_str + str(SL) if SL else plan

    return plan

def get_portf_data(exclude={}, port_filt_author='', port_filt_date_frm='',
                     port_filt_date_to='', port_filt_chn='', **kwargs ):
    fname_port = cfg['portfolio_names']['portfolio_fname']
    if not op.exists(fname_port):
        return [],[]
    try:
        data = pd.read_csv(fname_port,sep=",")
    except:
        data = pd.read_csv(fname_port,sep=",")

    data['Date'] = data['Date'].apply(lambda x: short_date(x))
    data['exit_plan']= data['exit_plan'].apply(lambda x: format_exitplan(x))
    data["isOpen"] = data["isOpen"].map({1:"Yes", 0:"No"})
    alerts = data[['STC1-Alerted', 'STC2-Alerted', 'STC3-Alerted']].sum(1)
    data["N Alerts"]= alerts.astype(int)
    data['Trader'] = data['Trader'].apply(lambda x: x.split('(')[0].split('#')[0])

    if exclude.get("live PnL", False):
        data =  get_live_quotes(data)

    for i in range(1,4):
        data[f'STC{i}-PnL'] = pd_col_str_frmt(data[f'STC{i}-PnL'])
        data[f'STC{i}-uQty'] = pd_col_str_frmt(data[f'STC{i}-uQty'])

    frm_cols = ['Price', 'Price-Alert', "Price-Current", 'uQty', 'filledQty', 'N Alerts', 
                "PnL", "$PnL","PnL-Alert", "$PnL-Alert","PnL-Current","$PnL-Current"]
    for cfrm in frm_cols:
        data[cfrm] = pd_col_str_frmt(data[cfrm])
    
    try:
        data = filter_data(data,exclude, port_filt_author, port_filt_date_frm,
                        port_filt_date_to, port_filt_chn)
    except:
        pass
    cols = ['isOpen', "PnL", "$PnL", 'Date', 'Symbol', 'Trader', 'BTO-Status', 'Price',
            'Price-Alert', "Price-Current", 'uQty', 'filledQty', 'N Alerts',"PnL-Alert",
            "$PnL-Alert","PnL-Current","$PnL-Current", "STC1-Price", "STC1-Price-Alerted",
            "STC1-Price-Current", 'STC1-PnL', 'STC1-Status','STC1-uQty','STC2-PnL',
            'STC2-Status', 'STC2-uQty', 'STC3-PnL', 'STC3-Status',  'STC3-uQty'
            ]
    data = data[cols]
    data.fillna("", inplace=True)
    header_list = data.columns.tolist()
    header_list = [d.replace('STC', '').replace("Price", "$") for d in header_list]
    
    if len(data):
        sumtotal = {c:"" for c in data.columns}
        for sumcol in ["PnL","PnL-Alert","PnL-Current",'STC1-PnL']:
            sumtotal[sumcol]= f'{data[sumcol].apply(lambda x: np.nan if x =="" else eval(x)).mean():.2f}'
        for sumcol in [ "$PnL","$PnL-Alert","$PnL-Current"]:
            sumtotal[sumcol]= f'{data[sumcol].apply(lambda x: np.nan if x =="" else eval(x)).sum():.2f}'
        sumtotal['Date'] = data.iloc[len(data)-1]['Date']
        sumtotal['Symbol'] = "Total Average"
        sumtotal['Trader'] = "Average"
        data = pd.concat([data, pd.DataFrame.from_records(sumtotal, index=[0])], ignore_index=True)
    
    data = data.values.tolist()
    return data, header_list

def get_tracker_data(exclude={}, track_filt_author='', track_filt_date_frm='',
                     track_filt_date_to='', track_filt_sym='', track_exc_author='',
                     track_exc_chn='',**kwargs ):
    fname_port = cfg['portfolio_names']['tracker_portfolio_name']
    if not op.exists(fname_port):
        return [],[]
    
    data = pd.read_csv(fname_port, sep=",")

    data['Date'] = data['Date'].apply(lambda x: short_date(x))
    data["isOpen"] = data["isOpen"].map({1:"Yes", 0:"No"})
    data["N Alerts"]= data['Avged']
    data['Trader'] = data['Trader'].apply(lambda x: x.split('(')[0].split('#')[0])
    
    if not exclude.get("live PnL", False):
        data =  get_live_quotes(data)
    
    frm_cols = ['Amount', 'N Alerts', 'STC-Amount','STC-Price','STC-Price-current','STC-PnL','STC-PnL-current',
                'STC-PnL$','STC-PnL$-current', 'Price', 'Price-current']
    for cfrm in frm_cols:
        data[cfrm] = pd_col_str_frmt(data[cfrm])
    
    try:
        data = filter_data(data,exclude, track_filt_author, track_filt_date_frm,
                        track_filt_date_to, track_filt_sym, track_exc_author, track_exc_chn)
    except:
        pass
    cols = ['isOpen','STC-PnL','STC-PnL-current', 'STC-PnL$','STC-PnL$-current', 'Date', 'Symbol', 'Trader', 'Price',
            "Price-current", 'Amount', 'N Alerts','STC-Amount','STC-Price','STC-Price-current','STC-Date','Channel'
            ]
    # data['Trader'] = data['Trader'].apply(lambda x: x.split('(')[0].split('#')[0])
    data = data[cols]
    data.fillna("", inplace=True)
    header_list = data.columns.tolist()
    header_list = [d.replace('STC', 'S') for d in header_list]
    if len(data):
        sumtotal = {c:None for c in data.columns}
        for sumcol in ['STC-PnL','STC-PnL-current']:
            sumtotal[sumcol]= f'{data[sumcol].apply(lambda x: np.nan if x =="" else eval(x)).mean():.2f}'
        for sumcol in ['STC-PnL$','STC-PnL$-current']:
            sumtotal[sumcol]= f'{data[sumcol].apply(lambda x: np.nan if x =="" else eval(x)).sum():.2f}'
        sumtotal['Date'] = data.iloc[len(data)-1]['Date']
        sumtotal['Symbol'] = "Total Average"
        sumtotal['Trader'] = track_filt_author if len(track_filt_author) else "Average"
        data = pd.concat([data, pd.DataFrame.from_records(sumtotal, index=[0])], ignore_index=True)
        data.fillna("", inplace=True)
    data = data.values.tolist()
    return data, header_list


def get_stats_data(exclude={}, stat_filt_author='', stat_filt_date_frm='',
                     stat_filt_date_to='', stat_filt_sym='', 
                     stat_max_trade_cap='', stat_max_qty='', trail_stop_perc='',
                     stat_exc_author='', stat_exc_chn='', stat_exc_sym='',
                     fname_port=None,
                     **kwargs ):
    if fname_port is None:
        fname_port = cfg['portfolio_names']['tracker_portfolio_name']
    if not op.exists(fname_port):
        return [],[]
    
    data = pd.read_csv(fname_port, sep=",")

    data['Date'] = data['Date'].apply(lambda x: datetime.strptime(x, "%Y-%m-%d %H:%M:%S.%f").strftime("%m/%d/%Y"))
    data["isOpen"] = data["isOpen"].map({1:"Yes", 0:"No"})
    data["N Alerts"]= data['Avged']
    data['Trader'] = data['Trader'].apply(lambda x: x.split('(')[0].split('#')[0])
    try:
        data = filter_data(data,exclude, stat_filt_author, stat_filt_date_frm,
                        stat_filt_date_to, stat_filt_sym, stat_exc_author, stat_exc_chn, stat_exc_sym)
    except:
        pass
    if stat_max_qty != "" or stat_max_trade_cap != "":
        if stat_max_qty != "" and stat_max_qty.isnumeric():
            stat_max_qty = int(stat_max_qty)
            print("stat_max_qty:", stat_max_qty)
            option_mult = (data['Asset'] == 'option').astype(int)
            option_mult[option_mult==1] = 100
            exceeds_cap = data['Amount'] > stat_max_qty
            data.loc[exceeds_cap, 'Amount'] = stat_max_qty

        if stat_max_trade_cap != "" and stat_max_trade_cap.isnumeric():
            stat_max_trade_cap = int(stat_max_trade_cap)
            print("stat_max_trade_cap:", stat_max_trade_cap)
            option_mult = (data['Asset'] == 'option').astype(int)
            option_mult[option_mult==1] = 100
            trade_value = data['Amount'] * data['Price'] * option_mult
            exceeds_cap = trade_value > stat_max_trade_cap
            data.loc[exceeds_cap, 'Amount'] = np.floor(stat_max_trade_cap / (data['Price'] * option_mult))
            data = data[data['Amount'] * data['Price'] * option_mult <= stat_max_trade_cap]
        mult =(data['Asset'] == 'option').astype(int) 
        mult[mult==0] = .01  # pnl already in %
        data['STC-PnL$'] = data['Amount'] * data['STC-PnL'] * data['Price'] * mult
        data['STC-PnL$-current'] = data['Amount'] * data['STC-PnL-current'] * data['Price-current'] * mult
        data['STC-PnL$'] = data['STC-PnL$'].round()
        data['STC-PnL$-current'] = data['STC-PnL$-current'].round()

    data['PnL diff'] = data['STC-PnL-current'] - data['STC-PnL']
    # Define the aggregation functions for each column
    agg_funcs = {'STC-PnL$': 'sum',
                 'STC-PnL$-current': 'sum',
                 'STC-PnL': 'mean',
                 'STC-PnL-current': 'mean',
                 'PnL diff' : "mean",
                 'Date': ['count', 'min', 'max']
                 }
    # Perform the groupby operation and apply the aggregation functions
    result_td = data.groupby('Trader').agg(agg_funcs)
    result_td = result_td.reset_index()
    
    result_ch = data.groupby('Channel').agg(agg_funcs)
    result_ch = result_ch.reset_index()
    result_ch = result_ch.rename({'Channel': 'Trader'}, axis=1)

    # make grand avg
    data["all"] = 1
    agg_values_all = data.groupby('all').agg(agg_funcs)    
    agg_values_all["Trader"] = "Total average"
    agg_values_all.loc[0, 'Trader'] = "Channels:"
    agg_values_all = agg_values_all.reset_index()

    result_td = pd.concat([result_td, agg_values_all, result_ch],axis=0, ignore_index=True)
    result_td.drop('all', axis=1, level=0, inplace=True)
    new_cols =[k for k in result_td.columns.get_level_values(0)]
    new_cols[-3] = "N Trades"
    new_cols[-2] = "Since"
    new_cols[-1] = "Last"
    result_td.columns = new_cols
    result_td = result_td.round(1)

    for cfrm in result_td.columns[1:-2]:
        result_td[cfrm] = pd_col_str_frmt(result_td[cfrm])
    result_td = result_td.fillna("")
    header_list = result_td.columns.tolist()
    header_list = [d.replace('STC-', '') for d in header_list]
    data = result_td.values.tolist()
    return data, header_list


def filter_data(data,exclude={}, track_filt_author='', track_filt_date_frm='',
                track_filt_date_to='', track_filt_sym='', track_exc_author='', track_exc_chn='', stat_exc_sym=''):
    if len(exclude):
        for k, v in exclude.items():
            if k == "Cancelled" and v and k in data.columns:
                data = data[data["BTO-Status"] !="CANCELED"]
            elif k == "Closed" and v:
                data = data[data["isOpen"] !="No"]
            elif k == "Open" and v:
                data = data[data["isOpen"] !="Yes"]
            elif k == "NegPnL" and v:
                col = "PnL" if "PnL" in data else 'STC-PnL'                
                pnl = data[col].apply(lambda x: np.nan if x =="" else eval(x) if isinstance(x, str) else x)     
                data = data[pnl > 0 ]
            elif k == "PosPnL" and v:
                col = "PnL" if "PnL" in data else 'STC-PnL' 
                pnl = data[col].apply(lambda x: np.nan if x =="" else eval(x) if isinstance(x, str) else x)
                data = data[pnl < 0 ]
            elif k == "stocks" and v:
                data = data[data["Asset"] !="stock"]
            elif k == "options" and v:
                data = data[data["Asset"] !="option"]

    if track_filt_author:
        msk = [x.strip() for x in track_filt_author.split(",")]
        data = data[data['Trader'].str.contains('|'.join(msk), case=False)]
    if track_filt_date_frm:
        data = data[data['Date'] >= track_filt_date_frm]
    if track_filt_date_to:
        data = data[data['Date'] <= track_filt_date_to]
    if track_filt_sym:
        msk = [x.strip() for x in track_filt_sym.split(",")]
        data = data[data['Symbol'].str.contains('|'.join(msk), case=False)]
    if track_exc_author:
        msk = [x.strip() for x in track_exc_author.split(",")]
        data = data[~data['Trader'].str.contains('|'.join(msk), case=False)]
    if track_exc_chn:
        msk = [x.strip() for x in track_exc_chn.split(",")]
        data = data[~data['Channel'].str.contains('|'.join(msk), case=False)]
    if stat_exc_sym:
        msk = [x.strip() for x in stat_exc_sym.split(",")]
        data = data[~data['Symbol'].str.contains('|'.join(msk), case=False)]
    return data

def get_live_quotes(portfolio):
    dir_quotes = cfg['general']['data_dir'] + '/live_quotes'
    track_symb = portfolio.loc[portfolio['isOpen']=='Yes', 'Symbol'].to_list()
    
    quotes_sym = {}
    for sym in track_symb: 
        fquote = f"{dir_quotes}/{sym}.csv"
        if not op.exists(fquote):
            continue
        
        with open(fquote, "r") as f:
            quotes = f.readlines()
        
        timestamp, quote = quotes[-1].split(',')  # in ms
        # quote_date = datetime.fromtimestamp(int(timestamp))
        # if (datetime.now() - quote_date).total_seconds() > 20:
        #     continue
        quotes_sym[sym] = float(quote.replace('\n', '').replace(' ', ''))
    
    for sym in quotes_sym:
        live_price = quotes_sym[sym]
        msk = (portfolio['Symbol']==sym) & (portfolio['isOpen']=='Yes')
        trades = portfolio.loc[msk]
        
        for _, trade in trades.iterrows():
            order= {
                "uQty": trade['Amount'] if pd.isnull(trade.get("STC-Amount")) else trade['Amount']- trade["STC-Amount"],
                "price": live_price,
                "Actual Cost": live_price,
                }                 
            stc_info = calc_stc_prices(trade, order)
            for k, v in stc_info.items():
                if k == "STC-Amount":
                    continue
                portfolio.loc[msk,k] = v
    return portfolio

def get_hist_msgs(filt_author='', filt_date_frm='', filt_date_to='',
                  filt_cont='', chan_name="option_alerts", **kwargs):
    # Provide arguments to filter

    data = pd.read_csv(op.join(cfg['general']['data_dir'] , f"{chan_name}_message_history.csv"))
    cols = ['Author', 'Date', 'Content', 'Parsed']
    data = data[cols]

    data = data[~data['Author'].str.contains('Xcapture')]
    data['Author'] = data['Author'].apply(lambda x: x.split('#')[0])
    data['Date'] = data['Date'].apply(lambda x: short_date(x))

    data = data.fillna("")
    if filt_author:
        data = data[data['Author'].str.contains(filt_author, case=False)]
    if filt_date_frm:
        data = data[data['Date'] > filt_date_frm]
    if filt_date_to:
        data = data[data['Date'] < filt_date_to]
    if filt_cont:
        data = data[data['Content'].str.contains(filt_cont, case=False)]

    header_list = data.columns.tolist()
    return data.values.tolist(), header_list

def get_acc_bals(bksession):
    acc_inf = bksession.get_account_info()
    # if grabing new access token return None, try again
    if acc_inf is None:  
        acc_inf = bksession.get_account_info()
    accnt= {"id" : acc_inf['securitiesAccount']['accountId'],
        "balance": acc_inf['securitiesAccount']['currentBalances'].get('liquidationValue', 0),
        "cash": acc_inf['securitiesAccount']['currentBalances'].get('cashBalance',0),
        "funds": acc_inf['securitiesAccount']['currentBalances'].get('availableFunds', 0),
        }
    return acc_inf, accnt

def get_pos(acc_inf):
    positions = acc_inf['securitiesAccount'].get('positions', [])
    pos_tab = []
    pos_headings = ["Sym", "Last", "price", "PnL_%", "PnL","Qty", "Val", "Cost"]
    for pos in positions:
        price= round(pos['averagePrice'], 2)
        # pnl = pos['currentDayProfitLoss']
        pnl_p = pos['currentDayProfitLossPercentage'] * 100
        uQty = pos['longQuantity']
        cost = round(price * uQty, 2)
        last = round(pos["marketValue"] / uQty, 2)
        sym = pos['instrument']['symbol']
        asset = pos['instrument']['assetType']
        val = pos["marketValue"]
        if pos['instrument']['assetType'] == "OPTION":
            cost = round(price * uQty * 100, 2)
            last = round(pos["marketValue"] / uQty, 2)/100

        pnl_t = round(val - cost, 2)
        if  cost == 0:
            pnl_p_t = 0
        else:
            pnl_p_t = round((val -cost)*100 / cost, 2)

        pos_vals = [sym, last, price, pnl_p_t ,pnl_t, uQty, val , cost]
        pos_tab.append(pos_vals)

    db = pd.DataFrame(data=pos_tab, columns=pos_headings)
    db = dataframe_num2str(db)
    return db.values.tolist(), pos_headings

def order_info_pars(ord_dic, ord_list):
    """ get info from order request
    :param ord_dic: dict with order info, from 'orderStrategies' or
    'childOrderStrategies'
    """
    ord_headings = ["Sym", "Act", "Strat", "Price/stp","Date", "Qty/fill", "Status", "ordId"]
    sing_ord = []

    price = ord_dic.get("price")
    stprice = ord_dic.get("stopPrice")
    # If SL order with no stopPrice
    if price is None:
        price = ord_dic.get("stopPrice")
        stprice = None
    price = f"{price}" if stprice == None else f"{price}/{stprice}"

    sing_ord.append(ord_dic['orderType'])
    sing_ord.append(price)
    # sing_ord.append(stprice)

    date = ord_dic['enteredTime'].split("+")[0]
    date = short_date(date, "%Y-%m-%dT%H:%M:%S")
    sing_ord.append(date)

    qty = round_int_flt(ord_dic['quantity'])
    fill = round_int_flt(ord_dic['filledQuantity'])
    qty_fll = f"{qty}/{fill}" if fill else f"{qty}"
    sing_ord.append(qty_fll)
    sing_ord.append(ord_dic['status'])
    sing_ord.append(str(ord_dic['orderId']))

    for leg in ord_dic['orderLegCollection']:
        sing_ord_c = sing_ord.copy()
        sing_ord_c.insert(0, leg['instrument']['symbol'])
        sing_ord_c.insert(1, leg["instruction"].replace('BUY_TO_OPEN', "BUY").replace('SELL_TO_CLOSE', "SELL"))
        ord_list.append(sing_ord_c)

    return ord_list, ord_headings

def get_orders(acc_inf):
    orders =acc_inf['securitiesAccount'].get('orderStrategies', [])
    if len(orders) == 0:
        return ["NoOrders"],  ["Sym", "Act", "Strat", "Price/stp","Date", "Qty/fill", "Status", "ordId"], []
    ord_tab, cols = [], []
    col = 0
    for ordr in orders:
        col = not col
        ord_type = ordr['orderStrategyType']
        if ord_type == "OCO":
            for chl in ordr['childOrderStrategies']:
                nlen = len(ord_tab)
                ord_tab, heads = order_info_pars(chl, ord_tab)
                nnlen = len(ord_tab) - nlen
                cols = cols + [col]*nnlen
        else:
            nlen = len(ord_tab)
            ord_tab, heads = order_info_pars(ordr, ord_tab)
            nnlen = len(ord_tab) - nlen
            cols = cols + [col]*nnlen

    db = pd.DataFrame(data=ord_tab, columns=heads)
    db = dataframe_num2str(db)
    return db.values.tolist(),heads, cols











import pandas as pd

from ..configurator import cfg
from . import BaseBroker

from td.orders import Order, OrderLeg
from td.client import TDClient

class TDA(BaseBroker):
    def __init__(self,account_n=0, accountId=None):
        self.account_n = account_n
        self.accountId = accountId

    def get_session(self, ):
        """Provide either:
            - account_n: indicating the orinal position from the accounts list
            (if only one account, it will be 0)
            - accountId: Number of the account

        auth is a dict with login info created with setup.py
        """
        if len(cfg['TDA']['client_id']) < 10:
            raise ValueError( "No TDA authentication file found, get credentials (in setup.py) to continue")
        # Create a new session, credentials path is required.
        self.session = TDClient(
            client_id=cfg['TDA']['client_id'],
            redirect_uri=cfg['TDA']['redirect_url'],
            credentials_path=cfg['TDA']['credentials_path']
        )

        # Login to the session
        success = self.session.login()

        if self.accountId is not None:
            self.session.accountId = self.accountId
        else:
            self.account_n = 0
            accounts_info = self.session.get_accounts(account="all")[self.account_n]
            self.accountId = accounts_info['securitiesAccount']['accountId']
        return success

    def send_order(self, new_order):
        order_response = self.session.place_order(account=self.accountId,
                                        order=new_order)
        order_id = order_response["order_id"]
        return order_response, order_id
    
    def cancel_order(self, order_id):
        return self.session.cancel_order(self.accountId, order_id)

    def get_order_info(self, order_id):  
        """
        order_status = 'REJECTED' | "FILLED" | "WORKING"
        """      
        order_info = self.session.get_orders(account=self.accountId, order_id=order_id)
        if order_info['orderStrategyType'] == "OCO":
            order_status = [
                order_info['childOrderStrategies'][0]['status'],
                order_info['childOrderStrategies'][1]['status']]
            if not order_status[0]==order_status[1]:
                print(f"OCO order status are different in ordID {order_id}: ",
                      f"{order_status[0]} vs {order_status[1]}")
            order_status = order_status[0]
        elif order_info['orderStrategyType'] in ['SINGLE', 'TRIGGER']:
            order_status = order_info['status']
        else:
            raise TypeError("Not sure type order. Check")
        return order_status, order_info

    def get_quotes(self, symbol:list):
        return self.session.get_quotes(instruments=symbol)

    def get_orders(self):
        pass

    def get_order_status(self, order_id):
        pass
    
    def get_account_info(self):
        acc_inf = self.session.get_accounts(self.accountId, ['orders','positions'])
        return acc_inf

    def get_positions_orders(self):
        acc_inf = self.get_account_info()

        df_pos = pd.DataFrame(columns=["symbol", "asset", "type", "Qty", "Avg Price", "PnL", "PnL %"])

        for pos in acc_inf['securitiesAccount']['positions']:
            long = True if pos["longQuantity"]>0 else False

            pos_inf = {
                "symbol":pos["instrument"]["symbol"],
                "asset":pos["instrument"]["assetType"],
                "type": "long" if  long else "short",
                "Avg Price": pos['averagePrice'],
                "PnL": pos["currentDayProfitLoss"],
                }
            pos_inf["Qty"] = int(pos[f"{pos_inf['type']}Quantity"])
            pos_inf["PnL %"] = pos_inf["PnL"]/(pos_inf["Avg Price"]*pos_inf["Qty"])
            df_pos =pd.concat([df_pos, pd.DataFrame.from_records(pos_inf, index=[0])], ignore_index=True)

        df_ordr = pd.DataFrame(columns=["symbol", "asset", "type", "Qty",
                                        "Price", "action"])
        return df_pos, df_ordr

    def make_BTO_lim_order(self, Symbol:str, uQty:int, price:float, strike=None, **kwarg):
        new_order=Order()
        new_order.order_strategy_type("TRIGGER")
        new_order.order_type("LIMIT")
        new_order.order_session('NORMAL')
        new_order.order_duration('GOOD_TILL_CANCEL')
        new_order.order_price(float(price))

        order_leg = OrderLeg()

        if strike is not None:
            order_leg.order_leg_instruction(instruction="BUY_TO_OPEN")
            order_leg.order_leg_asset(asset_type='OPTION', symbol=Symbol)
        else:
            order_leg.order_leg_instruction(instruction="BUY")
            order_leg.order_leg_asset(asset_type='EQUITY', symbol=Symbol)

        order_leg.order_leg_quantity(quantity=int(uQty))
        new_order.add_order_leg(order_leg=order_leg)
        return new_order


    def make_BTO_PT_SL_order(self, Symbol:str, uQty:int, price:float, PTs:list=None,
                            PTs_Qty:list=None, SL:float=None, SL_stop:float=None, **kwarg):
        new_order= self.make_BTO_lim_order(Symbol, uQty, price)

        if PTs == [None]:
            return new_order

        PTs_Qty = [ round(uQty * pqty) for pqty in PTs_Qty]
        for PT, pqty in zip(PTs, PTs_Qty):
            new_child_order = new_order.create_child_order_strategy()
            new_child_order = self.make_Lim_SL_order(Symbol, pqty, PT, SL, SL_stop, new_child_order)
            new_order.add_child_order_strategy(child_order_strategy=new_child_order)
        return new_order


    def make_Lim_SL_order(self, Symbol:str, uQty:int,  PT:float, SL:float, SL_stop:float=None, new_order=None, strike=None, **kwarg):
        if new_order is None:
            new_order = Order()
        new_order.order_strategy_type("OCO")

        child_order1 = new_order.create_child_order_strategy()
        child_order1.order_strategy_type("SINGLE")
        child_order1.order_type("LIMIT")
        child_order1.order_session('NORMAL')
        child_order1.order_duration('GOOD_TILL_CANCEL')
        child_order1.order_price(float(PT))

        child_order_leg = OrderLeg()

        child_order_leg.order_leg_quantity(quantity=uQty)
        if strike is not None:
            child_order_leg.order_leg_instruction(instruction="SELL_TO_CLOSE")
            child_order_leg.order_leg_asset(asset_type='OPTION', symbol=Symbol)
        else:
            child_order_leg.order_leg_instruction(instruction="SELL")
            child_order_leg.order_leg_asset(asset_type='EQUITY', symbol=Symbol)

        child_order1.add_order_leg(order_leg=child_order_leg)
        new_order.add_child_order_strategy(child_order_strategy=child_order1)

        child_order2 = new_order.create_child_order_strategy()
        child_order2.order_strategy_type("SINGLE")
        child_order2.order_session('NORMAL')
        child_order2.order_duration('GOOD_TILL_CANCEL')

        if SL_stop is not None:
            child_order2.order_type("STOP_LIMIT")
            child_order2.order_price(float(SL))
            child_order2.stop_price(float(SL_stop))
        else:
            child_order2.order_type("STOP")
            child_order2.stop_price(float(SL))

        child_order2.add_order_leg(order_leg=child_order_leg)
        new_order.add_child_order_strategy(child_order_strategy=child_order2)
        return new_order


    def make_STC_lim(self, Symbol:str, uQty:int, price:float, strike=None, **kwarg):
        new_order=Order()
        new_order.order_strategy_type("SINGLE")
        new_order.order_type("LIMIT")
        new_order.order_duration('GOOD_TILL_CANCEL')
        new_order.order_price(float(price))

        order_leg = OrderLeg()
        order_leg.order_leg_quantity(quantity=int(uQty))

        if strike is not None:
            new_order.order_session('NORMAL')
            order_leg.order_leg_instruction(instruction="SELL_TO_CLOSE")
            order_leg.order_leg_asset(asset_type='OPTION', symbol=Symbol)
        else:
            new_order.order_session('SEAMLESS')
            order_leg.order_leg_instruction(instruction="SELL")
            order_leg.order_leg_asset(asset_type='EQUITY', symbol=Symbol)
        new_order.add_order_leg(order_leg=order_leg)
        return new_order

    def make_STC_SL(self, Symbol:str, uQty:int, SL:float, strike=None,
                    SL_stop:float=None, new_order=Order(), **kwarg):
        new_order=Order()
        new_order.order_strategy_type("SINGLE")

        if SL_stop is not None:
            new_order.order_type("STOP_LIMIT")
            new_order.stop_price(float(SL_stop))
            new_order.order_price(float(SL))
        else:
            new_order.order_type("STOP")
            new_order.stop_price(float(SL))

        new_order.order_session('NORMAL')
        new_order.order_duration('GOOD_TILL_CANCEL')

        order_leg = OrderLeg()
        order_leg.order_leg_quantity(quantity=int(uQty))
        if strike is not None:
            order_leg.order_leg_instruction(instruction="SELL_TO_CLOSE")
            order_leg.order_leg_asset(asset_type='OPTION', symbol=Symbol)
        else:
            order_leg.order_leg_instruction(instruction="SELL")
            order_leg.order_leg_asset(asset_type='EQUITY', symbol=Symbol)
        new_order.add_order_leg(order_leg=order_leg)
        return new_order

    def make_STC_SL_trailstop(self, Symbol:str, uQty:int,  trail_stop_percent:float, new_order=None, **kwarg):
        if new_order is None:
            new_order = Order()
        new_order.order_strategy_type("SINGLE")
        new_order.order_strategy_type("SINGLE")
        new_order.order_type("TRAILING_STOP")
        new_order.order_session('NORMAL')
        new_order.order_duration('GOOD_TILL_CANCEL')
        new_order.stop_price_offset(trail_stop_percent)
        new_order.stop_price_link_type('PERCENT')
        new_order.stop_price_link_basis('BID')
        
        child_order_leg = OrderLeg()
        child_order_leg.order_leg_quantity(quantity=uQty)
        if len(Symbol.split("_")) > 1:
            child_order_leg.order_leg_instruction(instruction="SELL_TO_CLOSE")
            child_order_leg.order_leg_asset(asset_type='OPTION', symbol=Symbol)
        else:
            child_order_leg.order_leg_instruction(instruction="SELL")
            child_order_leg.order_leg_asset(asset_type='EQUITY', symbol=Symbol)
        new_order.add_order_leg(order_leg=child_order_leg)
        return new_order






#
#msft_quotes = TDSession.get_quotes(instruments=['MSFT'])
#
#
#optID= make_optionID('PLTR', '3/26', 27, 'C')
#opt_order = make_lim_option(optID, 1, 3.1)
#
#order_response = TDSession.place_order(account=self.TDsession.accountId,
#                                       order=opt_order)
#order_id = order_response["order_id"]
#order_info = TDSession.get_orders(account=accountId, order_id=order_id)
#order_status = order_info['status']
#
#
#TDSession.get_options_chain({"symbol":"PLTR", "contractType":"CALL", "strike":27,
#                             "fromDate":"2021-03-26", "toDate":"2021-03-26"})
#
#
#new_order = make_BTO_PT_SL_order("PLTR", 1, BTO=26.0, PT=30.0, SL=24.0, SL_stop=24.5)
#
#order_response = TDSession.place_order(account=accountId,
#                                       order=new_order)
#order_id = order_response["order_id"]
#order_info = TDSession.get_orders(account=accountId, order_id=order_id)
#order_status = order_info['childOrderStrategies'][0]['status']
#
#
#
#accounts_info = TDSession.get_accounts(account="all", fields=['orders'])[account_n]
#
#
#TDSession.cancel_order(self.TDsession.accountId, order_id)
#








import unittest
from DiscordAlertsTrader.configurator import cfg

class TestConfigurator(unittest.TestCase):

    def test_trailstop_val(self):
        trailstop_val = cfg['order_configs']['default_trailstop']
        self.assertTrue(trailstop_val == '' or eval(trailstop_val) > .9)

    def test_general_options(self):
        cfg['general'].getboolean('DO_BTO_TRADES')
        cfg['order_configs'].getboolean('sell_current_price')
        cfg['order_configs'].getboolean('auto_trade')

    def test_portfolio_names(self):
        for v in cfg['portfolio_names'].values():
            self.assertTrue(v.endswith(".csv"))

    def test_cfg_options_set(self):
        self.assertTrue(cfg['general']['BROKERAGE'] in ['', 'TDA', "webull", 'etrades'])
        self.assertTrue(cfg['order_configs']['default_bto_qty'] in ['buy_one', 'trade_capital'])
        
if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Apr 15 10:05:24 2021
@author: adonay
"""
from datetime import datetime
import pandas as pd
import os.path as op
import sys
import queue

# from pathlib import Path
# file = Path(__file__).resolve()
# parent, top = file.parent, file.parents[3]
# sys.path.append(str(top))
sys.path.append("/data/Dropbox (Partners HealthCare)/xtrades_repo/")
from place_order import (make_BTO_lim_order, send_order,
                         make_STC_lim, make_Lim_SL_order, make_STC_SL)
from alerts_trader import AlertTrader
from message_parser import parser_alerts, get_symb_prev_msg, combine_new_old_orders
import config_example as cfg



class TDSession_test():

    def __init__(self, order_log_fname="test_order_log.csv"):

        self.order_log_fname = order_log_fname
        if op.exists(self.order_log_fname):
            self.order_log = pd.read_csv(self.order_log_fname)
        else:
            self.order_log = pd.DataFrame(columns=[ "ordID",
                "symbol", "quantity", 'status', 'price', "closeTime", 'filledQuantity'])

        self.accountId = "19283746"
        self.quotes= pd.DataFrame(columns=["symbol", "askPrice", "bidPrice"])

    def make_quote(self, sym, ask, bid):
        inx_sym = self.quotes["symbol"].str.contains(sym)
        if any(inx_sym):
            self.quotes.loc[inx_sym,:] = [sym, ask, bid]
        else:
            self.quotes = self.quotes.append({"symbol":sym, "askPrice":ask,
                                          "bidPrice":bid}, ignore_index=True)

    def get_quotes(self, instruments):
        if isinstance(instruments, str): instruments = [instruments]

        quotes = {}
        for sym in instruments:
            sqt = self.quotes[self.quotes["symbol"] == sym]
            if sqt.size == 0:
                raise ValueError("quote not found, it needs to be created with self.make_quote")
            quotes[sym] = {'askPrice': sqt['askPrice'].iloc[0],
                           'bidPrice': sqt['bidPrice'].iloc[0]}
        return quotes


    def cancel_order(self, accountId, order_id):
        order_bool =  self.order_log['ordID'] == order_id
        assert sum(order_bool)==1

        order = self.order_log.loc[order_bool]
        if order["status"] == "WORKING":
            self.order_log.loc[order_bool, "status"] = "CANCELLED"
            print(f"Order {order_id} cancelled")

    def get_orders(self, account, order_id):
        # BTO
        msk = self.order_log["ordID"]  == order_id
        inx = msk[msk].index.values
        if not inx.size:
            raise ValueError ("ordID not found")
        elif inx.size >1:
            raise ValueError ("ordID found more than once")
        else:
            inx, = inx

        order_status = self.order_log.loc[inx, "status"]
        quantity = self.order_log.loc[inx, "quantity"]
        filledQuantity = self.order_log.loc[inx, "filledQuantity"]
        price = self.order_log.loc[inx, "price"]
        date = self.order_log.loc[inx, "closeTime"]

        order_info = {'orderLegCollection':[{'quantity': quantity}],
                      'quantity' :  quantity,
                      'filledQuantity': filledQuantity,
                      'status' : order_status,
                      'price' : price,
                      "closeTime" : date}

        if not pd.isnull(self.order_log.loc[inx, "OCO"]):
            order_info['orderStrategyType'] = "OCO"
            order_info['childOrderStrategies'] = [{"status": order_info['status']},
                                                  {"status": order_info['status']}]
        else:
            order_info['orderStrategyType'] = "SINGLE"

        return order_info


    def place_order(self, account, order):
        def make_ordID(n=1):
            # n = num ordIds
            ords = self.order_log["ordID"].to_numpy()
            inx = 0
            while True:
                if n > 1:
                    inxs = [inx + i for i in range(n)]
                    if not any(True for i in inxs if i in ords):
                        order_id = inx
                        break
                else:
                    if inx not in ords:
                        order_id = inx
                        break
                inx += 1
            return order_id

        order_response ={}
        # Strategy Order, e.g. OCO
        if order.template.get('orderStrategyType') =='OCO':

            chld_ord = order.child_order_strategies
            order_id = make_ordID(len(chld_ord))
            order_id_ret = order_id
            oco_inx = len(self.order_log)

            for chld in chld_ord.values():
                price = chld.get("price")
                stprice = chld.get("stopPrice")
                # If SL order with no stopPrice
                if price is None:
                    price = chld.get("stopPrice")
                    stprice = None

                ord_info = {
                    "ordID" : order_id,
                    "symbol" : chld['orderLegCollection'][0]["instrument"]['symbol'],
                    "quantity" : chld['orderLegCollection'][0]['quantity'],
                    "closeTime" : datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f"),
                    "price" : price,
                    "stopPrice": stprice,
                    "status" : "WORKING",
                    "filledQuantity" : 0,
                    "session" : chld["session"],
                    "orderType" : chld['orderType'],
                    "duration" :  chld["duration"],
                    "instruction" :chld['orderLegCollection'][0]["instruction"],
                    "strategyType" : chld['orderStrategyType'],
                    "OCO" : oco_inx
                   }
                # print(ord_info)
                self.order_log = self.order_log.append(ord_info, ignore_index=True)
                order_id += 1
            order_response["order_id"] = order_id_ret
            return order_response

        order_id = make_ordID(1)
        price = order.template.get('price')
        stprice = order.template.get('stopPrice')
        if price is None:
            price = order.template.get('stopPrice')
            stprice = None

        ord_info = {
            "ordID" : order_id,
            "symbol" : order.order_legs_collection['order_leg_1']['instrument']['symbol'],
            "quantity" : order.order_legs_collection['order_leg_1']['quantity'],
            "closeTime" : datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f"),
            "price" : price,
            "stopPrice": stprice,
            "status" : "WORKING",
            "filledQuantity" : 0,
            "session" :  order.template["session"],
            "orderType" :  order.template['orderType'],
            "duration" :   order.template["duration"],
            "instruction" : order.order_legs_collection['order_leg_1']["instruction"],
            "strategyType" : order.template['orderStrategyType'],
            "OCO" : None
           }
        # print(ord_info)

        self.order_log = self.order_log.append(ord_info, ignore_index=True)
        order_response["order_id"] = order_id
        return order_response

    def execute_orders(self):
        def fill_order(ix):
            self.order_log.loc[ix, "status"] = "FILLED"
            self.order_log.loc[ix, "filledQuantity"] = order["quantity"]
            close_OCO(ix)
            
        def close_OCO(ix):
            if not pd.isnull(self.order_log.loc[ix, "OCO"]):
                inx_oco = self.order_log.loc[ix, "OCO"]
                if inx_oco == ix: inx_oco +=1
                self.order_log.loc[inx_oco, "status"] = "CANCELLED"

        for ix, order in self.order_log.iterrows():
            if order["status"] != "WORKING":
                continue
            quote = self.get_quotes([order["symbol"]])[order["symbol"]]
            ask, bid = quote['askPrice'], quote['bidPrice']
            if order["instruction"] == "BUY" and  ask <= order["price"]:
                fill_order(ix)
            elif order["instruction"] == "SELL":
                if order["orderType"] == "LIMIT" and  bid >= order["price"]:
                    fill_order(ix)
                elif order["orderType"] == "STOP" and  bid <= order["price"]:
                    fill_order(ix)
                elif order["orderType"] == "STOP_LIMIT" and \
                    bid <= order["price"] and bid > order["stopPrice"]:
                    fill_order(ix)

if 0:
    tdt = TDSession_test()
    sym, ask, bid ='TSLA', 698, 696
    tdt.make_quote(sym,ask, bid)
    quote = tdt.get_quotes(sym)
    assert(list(quote.keys())[0] == sym)
    assert(quote[sym]["askPrice"] == ask)
    assert(quote[sym]["bidPrice"] == bid)


    Altrader = AlertTrader(portfolio_fname="trader_portfolio_simulated.csv",
                alerts_log_fname="trader_logger_simulated.csv",
                queue_prints=queue.Queue(maxsize=10),
                test_TDsession=tdt, update_portfolio=False)

    Altrader.price_now(sym, pflag=1)
    msg = "BTO TSLA @ 698 PT1 700 SL 690"
    pars, order =  parser_alerts(msg, "stock")
    order["Trader"] = "tester1"
    Altrader.new_trade_alert(order, pars,msg)

    tdt.make_quote('TSLA', 698, 699)
    tdt.execute_orders()
    Altrader.update_orders()

    order = {"Symbol" : "TSLA",
             "uQty" : 3,
             "price" : 690
             }
    new_order = make_BTO_lim_order(**order)
    ord_status, ordID = tdt.place_order(tdt.accountId, new_order)
    tdt.execute_orders()
    ord_status, ord_info = tdt.get_orders(tdt.accountId , ordID)

    tdt.make_quote("TSLA", 690, 686)
    tdt.execute_orders()
    ord_status, ord_info = tdt.get_orders(tdt.accountId , ordID)

    order = {"Symbol" : "TSLA",
             "uQty" : 3,
             "PT" : 695,
             "SL" : 680,
             "SL_stop" : 670
             }
    new_order = make_Lim_SL_order(**order)
    ord_status, ordID = tdt.place_order(tdt.accountId, new_order)

    tdt.make_quote("TSLA", 698, 696)
    tdt.execute_orders()
    ord_status, ord_info = tdt.get_orders(tdt.accountId, ordID)


    order = {"Symbol" : "STC-lim",
             "uQty" : 3,
             "price" : 3.5,
             "PT" : 99,
             "SL" : 99,
             "SL_stop" : 99
              }

    new_order = make_STC_lim(**order)
    tdt.place_order(tdt.accountId, new_order)


    order = {"Symbol" : "STC-SL",
             "uQty" : 3,
             "price" : 99,
             "PT" : 99,
             "SL" : 1.4,
             "SL_stop" : 1.6
              }
    new_order =  make_STC_SL(**order)
    tdt.place_order(tdt.accountId, new_order)


    order = {"Symbol" : "STC-SL",
             "uQty" : 3,
             "price" : 99,
             "PT" : 99,
             "SL" : 1.4,
              }
    new_order =  make_STC_SL(**order)
    tdt.place_order(tdt.accountId, new_order)




import re
from datetime import timezone
import pandas as pd

def bot_msgs(message):
    time_strf = "%Y-%m-%d %H:%M:%S.%f"
    # only respond to ch_id
    ch_id = 862187203711074317
    if message.channel.id == ch_id:
        msg_date = message.created_at.replace(tzinfo=timezone.utc).astimezone(tz=None)
        msg_date_f = msg_date.strftime(time_strf)    

        try: cont = message.embeds[0].description
        except IndexError: 
            print("no embeddings", message.content)
            return
        
        # Get algo name as author        
        algo_names = [('algo_1', ":black_joker: Scalp Opening :black_joker:"),
        ('algo_2', ":spy: Scalp Opening :spy:"),
        ('algo_3', ":man_mage: Scalp Opening :man_mage:")]
        author = None
        for name, patt in algo_names:
            if patt in cont:
                author = name
                break
        if author is None:
            return
        
        # get alert info
        alert =  parse_alert(cont)
        msg = pd.Series({'AuthorID': message.author.id,
                        'Author': author,
                        'Date': msg_date_f, 
                        'Content': alert,
                        'Channel': "fend_bot"
                            })
        return msg


def parse_alert(msg):
    id_exp = r"Eyeing:(\w+)_(\d+)_([\d\.]+)_(C|P)"
    match = re.search(id_exp, msg)
    if match is not None:
        symbol, expdate, strike, call_put = match.groups()
        current_mark = re.search(r"Current Mark:([\d.]+)", msg).group(1)
        alert_string = f"BTO {symbol} {strike}{call_put} {expdate[:2]}/{expdate[2:4]} @ {current_mark}"
        return alert_string


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Feb 27 09:26:06 2021

@author: adonay
"""
import re
import os.path as op
import numpy as np
import pandas as pd
from datetime import datetime, date
import time
import threading
import queue
from colorama import Fore, Back

from DiscordAlertsTrader.configurator import cfg
from DiscordAlertsTrader.message_parser import parse_exit_plan, set_exit_price_type


def find_last_trade(order, trades_log, open_only=True):
    trades_authr = trades_log["Trader"] == order["Trader"]
    trades_log = trades_log.loc[trades_authr]

    msk_ticker = trades_log["Symbol"].str.match(f"{order['Symbol']}$")

    # Order ticker without dates and strike
    if sum(msk_ticker) == 0 and order['asset'] == 'option':
        trades_log = trades_log[trades_log["Asset"] == "option"]
        trade_symb = trades_log["Symbol"].apply(lambda x: x.split("_")[0])

        msk_ticker = trade_symb.str.match(f"{order['Symbol']}$")

    if sum(msk_ticker) == 1:
        last_trade, = trades_log[msk_ticker].index.values
    # Either take open trade or last
    elif sum(msk_ticker) > 1:
        open_trade = trades_log.loc[msk_ticker, "isOpen"]
        if open_trade.sum() == 1:
            last_trade, = open_trade.index[open_trade==1]
        elif open_trade.sum() > 1:
            # raise ValueError ("Trade with more than one open position")
            last_trade = open_trade.index[open_trade==1][-1]
        elif open_trade.sum() == 0:
            last_trade = open_trade.index[-1]
    else:
        return None, 0

    isOpen = trades_log.loc[last_trade, 'isOpen']

    if open_only and isOpen == 0:
        return None, 0
    else:
        return last_trade, isOpen


class AlertsTrader():
    def __init__(self,
                 brokerage,
                 portfolio_fname=cfg['portfolio_names']['portfolio_fname'] ,
                 alerts_log_fname=cfg['portfolio_names']['alerts_log_fname'],
                 queue_prints=queue.Queue(maxsize=10),
                 update_portfolio=True):
        self.bksession = brokerage
        self.portfolio_fname = portfolio_fname
        self.alerts_log_fname = alerts_log_fname
        self.queue_prints = queue_prints
        
        # load port and log
        if op.exists(self.portfolio_fname):
            self.portfolio = pd.read_csv(self.portfolio_fname)
        else:
            self.portfolio = pd.DataFrame(columns=cfg["col_names"]['portfolio'].split(",") )
            self.portfolio.to_csv(self.portfolio_fname, index=False)
        if op.exists(self.alerts_log_fname):
            self.alerts_log = pd.read_csv(self.alerts_log_fname)
        else:
            self.alerts_log = pd.DataFrame(columns=cfg["col_names"]['alerts_log'].split(","))
            self.alerts_log.to_csv(self.alerts_log_fname, index=False)

        self.update_portfolio = update_portfolio
        self.update_paused = False
        if update_portfolio:
            # first do a synch, then thread it
            self.update_orders()
            self.activate_trade_updater()

    def activate_trade_updater(self, refresh_rate=30):
        self.update_portfolio = True
        self.updater = threading.Thread(target=self.trade_updater, args=[refresh_rate])
        self.updater.start()
        self.queue_prints.put([f"Updating portfolio orders every {refresh_rate} secs", "", "green"])
        print(Back.GREEN + f"Updating portfolio orders every {refresh_rate} secs")

    def trade_updater_reset(self, refresh_rate=30):
        """ Will stop threding updater and restart.
        To avoid delays or parallel updatings. """
        self.update_portfolio = False
        time.sleep(refresh_rate)
        self.activate_trade_updater(refresh_rate)

    def _stop(self):
        "Lazy trick to stop updater threading"
        self.update_portfolio = False

    def trade_updater(self, refresh_rate=30):
        while self.update_portfolio is True:
            if self.update_paused is False:
                try:
                    self.update_orders()
                except Exception as ex:
                    str_msg = f"Error raised during port update, trying again later. Error: {ex}"
                    print(Back.RED + str_msg)
                    self.queue_prints.put([str_msg, "", "red"])
            if self.update_portfolio:
                time.sleep(refresh_rate)
        str_msg = "Closed portfolio updater"
        print(Back.GREEN + str_msg)
        self.queue_prints.put([str_msg, "", "green"])

    def save_logs(self, csvs=["port", "alert"]):
        if "port" in csvs:
            self.portfolio.to_csv(self.portfolio_fname, index=False)
        if "alert" in csvs:
            self.alerts_log.to_csv(self.alerts_log_fname, index=False)

    def order_to_pars(self, order):
        pars_str = f"{order['action']} {order['Symbol']} @{order['price']}"
        if {order['action']} == "BTO":
            for i in range(1, 4):
                pt = f"PT{i}"
                if pt in order.keys() and order[pt] is not None:
                    pars_str = pars_str + f" {pt}: {order[pt]}"
                if "SL" in order.keys() and order["SL"] is not None:
                    pars_str = pars_str + f" SL: {order['SL']}"
        elif {order['action']} == "STC":
            pars_str = pars_str + f" Qty:{order['uQty']}({int(order['xQty']*100)}%)"
        return pars_str

    def price_now(self, symbol, price_type="BTO", pflag=0):
        if price_type in ["BTO", "BTC"]:
            ptype = 'askPrice'
        else:
            ptype= 'bidPrice'
        try:
            resp = self.bksession.get_quotes([symbol])
            if resp is None or len(resp) == 0 or symbol not in resp.keys() or resp[symbol].get('description' ) == 'Symbol not found' :
                str_msg =  f"{symbol} not found during price quote"
                print (Back.RED + str_msg)
                self.queue_prints.put([str_msg, "", "red"])
                quote = -1
            else:
                quote = resp[symbol][ptype]
        except KeyError as e:
            str_msg= f"price_now error for symbol {symbol}: {e}.\n Try again later"
            print (Back.RED + str_msg)
            self.queue_prints.put([str_msg, "", "red"])
            quote = self.bksession.get_quotes(symbol)[symbol][ptype]
        if pflag:
            return quote
        else:
            return "CURRENTLY @%.2f"% quote

    def confirm_and_send(self, order, pars, order_funct):
            resp, order, ord_chngd = self.notify_alert(order, pars)
            if resp in ["yes", "y"]:
                try:
                    ord_resp, ord_id = self.bksession.send_order(order_funct(**order))
                except Exception as e:
                    str_msg = f"Error in order {e}"
                    print(Back.GREEN + str_msg)
                    self.queue_prints.put([str_msg, "", "red"])
                    return None, None, order, None
                
                if ord_resp is None:
                    raise("Something wrong with order response")

                str_msg = f"Sent order {pars}"
                print(Back.GREEN + str_msg)
                color = "green" if order['action'] == "BTO" else "yellow"
                self.queue_prints.put([str_msg, "", color])
                return ord_resp, ord_id, order, ord_chngd

            elif resp in ["no", "n"]:
                return None, None, order, None

    def notify_alert(self, order, pars):
        price_now = self.price_now
        symb = order['Symbol']
        ord_ori = order.copy()
        pars_ori = pars
        act = order['action']

        while True:
            # If symbol not found, quote val returned is -1
            if price_now(symb, act, 1 ) == -1:
                return "no", order, False

            current_price = price_now(symb, act, 1 )
            order["price_current"] = current_price
            pdiff = (current_price - ord_ori['price'])/ord_ori['price']
            pdiff = round(pdiff*100,1)

            question = f"{pars_ori} {price_now(symb, act)}"
            if cfg['order_configs'].getboolean('sell_current_price'):
                if pdiff < eval(cfg['order_configs']['max_price_diff'])[order["asset"]]:
                    order['price'] = price_now(symb, act, 1)
                    pars = self.order_to_pars(order)
                    question += f"\n new price: {pars}"
                else:
                    if cfg['order_configs'].getboolean('auto_trade') is True and order['action'] == "BTO":
                        str_msg = f"BTO alert price diff too high: {pdiff}% at {current_price}, keeping original price of {ord_ori['price']}"
                        print(Back.GREEN + str_msg)
                        self.queue_prints.put([str_msg, "", "green"])

            if cfg['order_configs'].getboolean('auto_trade') is True:
                if cfg['general'].getboolean('DO_BTO_TRADES') is False and order['action'] == "BTO":
                    str_msg = f"BTO not accepted by config options: DO_BTO_TRADES = False"
                    print(Back.GREEN + str_msg)
                    self.queue_prints.put([str_msg, "", "green"])
                    return "no", order, False
                elif order['action'] == "BTO":
                    price = order['price']
                    if price == 0:
                        str_msg = f"Order not accepted price is 0"
                        print(Back.GREEN + str_msg)
                        self.queue_prints.put([str_msg, "", "red"])
                        return "no", order, False
                    price = price*100 if order["asset"] == "option" else price
                    max_trade_val = float(cfg['order_configs']['max_trade_capital'])

                    if 'uQty' not in order.keys() or order['uQty'] is None:
                        if cfg['order_configs']['default_bto_qty'] == "buy_one":
                            order['uQty'] = 1                    
                        elif cfg['order_configs']['default_bto_qty'] == "trade_capital":
                            order['uQty'] =  int(max(round(float(cfg['order_configs']['trade_capital'])/price), 1))

                    if price * order['uQty'] > max_trade_val:
                        uQty_ori = order['uQty']
                        order['uQty'] =  int(max(max_trade_val//price, 1))
                        if price * order['uQty'] <= max_trade_val:
                            str_msg = f"BTO trade exeeded max_trade_capital of ${max_trade_val}, order quantity reduced to {order['uQty']} from {uQty_ori}"
                            print(Back.GREEN + str_msg)
                            self.queue_prints.put([str_msg, "", "green"])
                        else:
                            str_msg = f"cancelled BTO: trade exeeded max_trade_capital of ${max_trade_val}"
                            print(Back.RED + str_msg)
                            self.queue_prints.put([str_msg, "", "red"])
                            return "no", order, False
                return "yes", order, False

            # Manual trade 
            resp = input(Back.RED  + question + "\n Make trade? (y, n or (c)hange) \n").lower()

            if resp in [ "c", "change", "y", "yes"] and 'uQty' not in order.keys():
                order['uQty'] = int(input("Order qty not available." +
                                          f" How many units to buy? {price_now(symb, act)} \n"))

            if resp in [ "c", "change"]:
                new_order = order.copy()
                new_order['price'] = float(input(f"Change price @{order['price']}" +
                                        f" {price_now(symb, act)}? Leave blank if NO \n")
                                      or order['price'])

                if order['action'] == 'BTO':
                    PTs = [order[f'PT{i}'] for i in range(1,4)]
                    PTs = eval(input(f"Change PTs @{PTs} {price_now(symb, act)}? \
                                      Leave blank if NO, respond eg [1, 2, None] \n")
                                      or str(PTs))

                    new_n = len([i for i in PTs if i is not None ])
                    if new_n != order['n_PTs']:
                        new_order['n_PTs']= new_n
                        new_order['PTs_Qty'] = [round(1/new_n,2) for i in range(new_n)]
                        new_order['PTs_Qty'][-1] = new_order['PTs_Qty'][-1] + (1- sum(new_order['PTs_Qty']))

                    new_order["SL"] = (input(f"Change SL @{order['SL']} {price_now(symb, act)}?"+
                                             " Leave blank if NO \n")
                                       or order['SL'])

                    new_order["SL"] = eval(new_order["SL"]) if isinstance(new_order["SL"], str) else new_order["SL"]
                order = new_order
                pars = self.order_to_pars(order)
            else :
                break
        ord_chngd = ord_ori != order
        return resp, order, ord_chngd


    def close_open_exit_orders(self, open_trade, STCn=range(1,4)):
        # close STCn waiting orders
        position = self.portfolio.iloc[open_trade]
        if type(STCn) == int: STCn = [STCn]

        for i in STCn:
            if pd.isnull(position[ f"STC{i}-ordID"]):
                continue

            order_id =  int(position[ f"STC{i}-ordID"])
            ord_stat, _ = self.get_order_info(order_id)

            if ord_stat not in ["FILLED", "EXECUTED", 'CANCELED','CANCEL_REQUESTED','REJECTED', 'EXPIRED']:
                print(Back.GREEN + f"Cancelling {position['Symbol']} STC{i}")
                self.queue_prints.put([f"Cancelling {position['Symbol']} STC{i}", "", "green"])
                _ = self.bksession.cancel_order(order_id)

                self.portfolio.loc[open_trade, f"STC{i}-Status"] = np.nan
                self.portfolio.loc[open_trade, f"STC{i}-ordID"] = np.nan
                self.save_logs("port")


    def get_order_info(self, order_id):
        try:
             order_status, order_info = self.bksession.get_order_info(order_id)
             return order_status, order_info
        except Exception as ex:
            print(f"Caught Error in order info, skipping order info retr. Error: {ex}")
            self.queue_prints.put([f"Caught Error, skipping order info retr. Error: {ex}", "", "red"])
            return None, None


    ######################################################################
    # ALERT TRADER
    ######################################################################

    def new_trade_alert(self, order:dict, pars:str, msg):
        """ get order from ```parser_alerts``` """

        open_trade, isOpen = find_last_trade(order, self.portfolio)

        time_strf = "%Y-%m-%d %H:%M:%S.%f"
        date = datetime.now().strftime(time_strf)

        log_alert = {"Date": date,
                     "Symbol": order['Symbol'],
                     "Trader" : order['Trader'],
                     "parsed" : pars,
                     "msg": msg
                     }

        if order['action'] == "ExitUpdate" and isOpen:
            #TODO : REMOVE
            # Pause updater to avoid overlapping
            self.update_paused = True

            old_plan = self.portfolio.loc[open_trade, "exit_plan"]
            new_plan = parse_exit_plan(order)

            # check if asset if price stock or contract
            if self.portfolio.loc[open_trade, "Asset"] == "option":
                new_plan["price"] = self.portfolio.loc[open_trade, "Price"]
                sym_inf = self.portfolio.loc[open_trade, "Symbol"].split("_")[1]
                strike = re.split("C|P", sym_inf)[1]
                new_plan["strike"] = strike + "C"
                for i in range(1,4):
                    exit_price = new_plan.get(f"PT{i}")
                    if exit_price is not None:
                        new_plan[f"PT{i}"] = set_exit_price_type(exit_price, new_plan)
                    if new_plan.get("SL"):
                        new_plan[f"SL"] = set_exit_price_type(new_plan.get("SL"), new_plan)
                _ = [new_plan.pop(k) for k in ['price', 'strike']]

            # Update PT is already STCn
            istc = None
            for i in range(1,3):
                if not pd.isnull(self.portfolio.loc[open_trade, f"STC{i}-Alerted"]):
                    istc = i+1
            if istc is not None and any(["PT" in k for k in new_plan.keys()]):
                new_plan_c = new_plan.copy()
                for i in range(1,4):
                    if new_plan.get(f"PT{i}"):
                        del new_plan_c[f"PT{i}"]
                        new_plan_c[f"PT{istc}"] = new_plan[f"PT{i}"]
                        # Cancel orders previous plan if any
                        self.close_open_exit_orders(open_trade, istc)
                new_plan = new_plan

            renew_plan = eval(old_plan)
            if renew_plan is not None or renew_plan != {}:
                for k in new_plan.keys():
                    renew_plan[k] = new_plan[k]
            else:
                renew_plan = new_plan

            self.portfolio.loc[open_trade, "exit_plan"] = str(renew_plan)
            self.update_paused = False

            log_alert['action'] = "ExitUpdate"
            self.save_logs()

            symb = self.portfolio.loc[open_trade, "Symbol"]
            print(Back.GREEN + f"Updated {symb} exit plan from :{old_plan} to {renew_plan}")
            self.queue_prints.put([f"Updated {symb} exit plan from :{old_plan} to {renew_plan}", "", "green"])
            return

        if not isOpen and order["action"] == "BTO":
            alert_price = order['price']

            order_response, order_id, order, _ = self.confirm_and_send(order, pars, self.bksession.make_BTO_lim_order)
            self.save_logs("port")
            if order_response is None:  #Assume trade not accepted
                log_alert['action'] = "BTO-notAccepted"                
                self.alerts_log = pd.concat([self.alerts_log, pd.DataFrame.from_records(log_alert, index=[0])], ignore_index=True)
                self.save_logs(["alert"])
                str_msg = "BTO not accepted by user, order response is none"
                print(Back.GREEN + str_msg)
                self.queue_prints.put([str_msg, "", "green"])
                return

            order_status, order_info = self.get_order_info(order_id)
            if order_status == 'REJECTED':                
                log_alert['action'] = "REJECTED"
                self.alerts_log = pd.concat([self.alerts_log, pd.DataFrame.from_records(log_alert, index=[0])], ignore_index=True)
                self.save_logs(["alert"])
                print(Back.GREEN + "BTO REJECTED")
                self.queue_prints.put(["BTO REJECTED", "", "green"])
                return
            
            exit_plan = parse_exit_plan(order)
            new_trade = {"Date": date,
                         "Symbol": order['Symbol'],
                         'isOpen': 1,
                         'BTO-Status' : order_status,
                         "uQty": order_info['quantity'],
                         "Asset" : order["asset"],
                         "Type" : "BTO",
                         "Price" : order_info["price"],
                         "Price-Alert" : alert_price,
                         "Price-Current": order["price_current"],
                         "ordID" : order_id,
                         "exit_plan" : str(exit_plan),
                         "Trader" : order['Trader'],
                         "Risk" : order['risk'],
                         "SL_mental" : order.get("SL_mental")
                         }

            self.portfolio = pd.concat([self.portfolio, pd.DataFrame.from_records(new_trade, index=[0])], ignore_index=True)
            
            if order_status in ["FILLED", "EXECUTED"]:
                ot, _ = find_last_trade(order, self.portfolio)
                self.portfolio.loc[ot, "Price"] = order_info['price']
                self.portfolio.loc[ot, "filledQty"] = order_info['filledQuantity']
            print(Back.GREEN + f"BTO {order['Symbol']} executed @ {order_info['price']}. Status: {order_status}")
            self.queue_prints.put([f"BTO {order['Symbol']} executed. Status: {order_status}", "", "green"])

            #Log portfolio, trades_log
            log_alert['action'] = "BTO"
            log_alert["portfolio_idx"] = len(self.portfolio) - 1
            self.alerts_log = pd.concat([self.alerts_log, pd.DataFrame.from_records(log_alert, index=[0])], ignore_index=True)
            self.save_logs()

        elif order["action"] == "BTO" and order['avg'] is not None:
            # if PT in order: cancel previous and make_BTO_lim_rder
            # else : BTO
            alert_price = order['price']
            order_response, order_id, order, _ = self.confirm_and_send(order, pars,
                                                                       self.bksession.make_BTO_lim_order)
            self.save_logs("port")
            if order_response is None:  #Assume trade not accepted
                log_alert['action'] = "BTO-Avg-notAccepted"
                self.alerts_log = pd.concat([self.alerts_log, pd.DataFrame.from_records(log_alert, index=[0])], ignore_index=True)
                self.save_logs(["alert"])
                print(Back.GREEN + "BTO avg not accepted by user")
                self.queue_prints.put(["BTO avg not accepted by user", "", "green"])
                return

            order_status, order_info = self.get_order_info(order_id)
            self.portfolio.loc[open_trade,'BTO-avg-Status'] = order_status
            self.portfolio.loc[open_trade,"ordID"] += f',{order_id}'

            if pd.isnull(self.portfolio.loc[open_trade, "Avged"]):
                self.portfolio.loc[open_trade, "Avged"] = 1
                self.portfolio.loc[open_trade, "Avged-prices-alert"] = alert_price
                self.portfolio.loc[open_trade, "Avged-prices"] = order_info["price"]
                self.portfolio.loc[open_trade, "Avged-uQty"] = order_info['quantity']

            else:
                self.portfolio.loc[open_trade, "Avged"] += 1
                al_pr = self.portfolio.loc[open_trade, "Avged-prices-alert"]
                av_pr = self.portfolio.loc[open_trade, "Avged-prices"]
                av_qt = self.portfolio.loc[open_trade, "Avged-uQty"]
                self.portfolio.loc[open_trade, "Avged-prices-alert"] = f"{al_pr},{alert_price}"
                self.portfolio.loc[open_trade, "Avged-prices"] = f"{av_pr},{order_info['price']}"
                self.portfolio.loc[open_trade, "Avged-uQty"] = f"{av_qt},{order_info['quantity']}"

            avg = self.portfolio.loc[open_trade, "Avged"]

            self.portfolio.loc[open_trade, "uQty"] += order_info['quantity']
            if  order_status in ["FILLED", "EXECUTED"]:
                self.portfolio.loc[open_trade, "filledQty"] += order_info['filledQuantity']
                self.close_open_exit_orders(open_trade)
            str_msg =  f"BTO {avg} th AVG, {order['Symbol']} executed. Status: {order_status}"
            print(Back.GREEN + str_msg)
            self.queue_prints.put([str_msg, "", "green"])

            #Log portfolio, trades_log
            log_alert['action'] = "BTO-avg"
            log_alert["portfolio_idx"] = len(self.portfolio) - 1
            self.alerts_log = pd.concat([self.alerts_log, pd.DataFrame.from_records(log_alert, index=[0])], ignore_index=True)
            self.save_logs()

        elif order["action"] == "BTO":
            str_act = "Repeated BTO"
            log_alert['action'] = "BTO-Null-Repeated"
            self.alerts_log = pd.concat([self.alerts_log, pd.DataFrame.from_records(log_alert, index=[0])], ignore_index=True)
            self.save_logs(["alert"])
            print(Back.RED + str_act)
            self.queue_prints.put([str_act, "", "red"])

        elif order["action"] == "STC" and isOpen == 0:
            open_trade, _ = find_last_trade(order, self.portfolio, open_only=False)
            if open_trade is None:
                log_alert['action'] = str_msg = f"STC-alerted without open position"
                self.alerts_log = pd.concat([self.alerts_log, pd.DataFrame.from_records(log_alert, index=[0])], ignore_index=True)
                self.save_logs()
                if cfg['general'].getboolean('DO_BTO_TRADES'):
                    print(Back.GREEN + str_msg)
                    self.queue_prints.put([str_msg, "", "green"])
                return

            position = self.portfolio.iloc[open_trade]
            # Check if closed position was not alerted
            for i in range(1,4):
                STC = f"STC{i}"
                if pd.isnull(position[f"{STC}-Alerted"]):
                    self.portfolio.loc[open_trade, f"{STC}-Alerted"] = 1
                    # If alerted and already sold
                    if not pd.isnull(position[ f"{STC}-Price"]):
                        print(Back.RED + "Position already closed")
                        self.queue_prints.put(["Position already closed", "", "red"])
                        log_alert['action'] = f"{STC}-alerterdAfterClose"
                        log_alert["portfolio_idx"] = open_trade
                        self.alerts_log = pd.concat([self.alerts_log, pd.DataFrame.from_records(log_alert, index=[0])], ignore_index=True)
                        self.save_logs()
                    return

            str_act = "STC without BTO, maybe alredy sold"
            log_alert['action'] = "STC-Null-notOpen"
            self.alerts_log = pd.concat([self.alerts_log, pd.DataFrame.from_records(log_alert, index=[0])], ignore_index=True)
            self.save_logs(["alert"])
            print(Back.RED + str_act)
            self.queue_prints.put([str_act, "", "red"])

        elif order["action"] == "STC":
            position = self.portfolio.iloc[open_trade]
            if order.get("amnt_left"):
                order, changed = amnt_left(order, position)
                print(Back.GREEN + f"Based on alerted amnt left, Updated order: " +
                      f"xQty: {order['xQty']} and uQty: {order['uQty']}")
                self.queue_prints.put([f"Based on alerted amnt left, Updated order: " +
                      f"xQty: {order['xQty']} and uQty: {order['uQty']}", "", "green"])

            # check if position already alerted and closed
            for i in range(1,4):
                STC = f"STC{i}"

                # If not alerted, mark it
                if pd.isnull(position[f"{STC}-Alerted"]):
                    self.portfolio.loc[open_trade, f"{STC}-Alerted"] = 1 
                    self.portfolio.loc[open_trade, STC + "-Price-Alerted"] = order["price"]
                    # If alerted and already sold
                    if not pd.isnull(position[ f"{STC}-Price"]):
                        print(Back.GREEN + "Already sold")
                        self.queue_prints.put(["Already sold", "", "green"])

                        log_alert['action'] = f"{STC}-DoneBefore"
                        log_alert["portfolio_idx"] = open_trade
                        self.alerts_log = pd.concat([self.alerts_log, pd.DataFrame.from_records(log_alert, index=[0])], ignore_index=True)
                        self.save_logs(["alert"])

                        if order['xQty'] != 1:  # if partial and sold, leave
                            return
                    break

            else:
                str_STC = "How many STC already?"
                print (Back.RED + str_STC)
                self.queue_prints.put([str_STC, "", "red"])
                log_alert['action'] = "STC-TooMany"
                log_alert["portfolio_idx"] = open_trade
                self.alerts_log = pd.concat([self.alerts_log, pd.DataFrame.from_records(log_alert, index=[0])], ignore_index=True)
                self.save_logs(["alert"])
                return

            qty_bought = position["filledQty"]

            if position["BTO-Status"] in ["CANCELED", "REJECTED", "EXPIRED", "CANCEL_REQUESTED"]:
                log_alert['action'] = "Trade-already cancelled"
                log_alert["portfolio_idx"] = open_trade
                self.alerts_log = pd.concat([self.alerts_log, pd.DataFrame.from_records(log_alert, index=[0])], ignore_index=True)
                self.save_logs(["alert"])
                return

            # Close position of STC All or STC SL
            if qty_bought == 0 and order['xQty'] == 1:

                order_id = position['ordID']
                _ = self.bksession.cancel_order(order_id)

                self.portfolio.loc[open_trade, "isOpen"] = 0

                order_status, _ =  self.get_order_info(order_id)
                self.portfolio.loc[open_trade, "BTO-Status"] = order_status

                print(Back.GREEN + f"Order Cancelled {order['Symbol']}, closed before fill")
                self.queue_prints.put([f"Order Cancelled {order['Symbol']}, closed before fill", "", "green"])

                log_alert['action'] = "STC-ClosedBeforeFill"
                log_alert["portfolio_idx"] = open_trade
                self.save_logs()
                # self.activate_trade_updater()
                return
            # Set STC as exit plan
            elif qty_bought == 0:
                exit_plan = eval(self.portfolio.loc[open_trade, "exit_plan"])
                exit_plan[f"PT{STC[-1]}"] = order["price"]
                self.portfolio.loc[open_trade, "exit_plan"] = str(exit_plan)
                str_msg = f"Exit Plan {order['Symbol']} updated, with PT{STC[-1]}: {order['price']}"
                print(Back.GREEN + str_msg)
                self.queue_prints.put([str_msg,"", "green"])
                log_alert['action'] = "STC-partial-BeforeFill-ExUp"
                log_alert["portfolio_idx"] = open_trade
                return

            qty_sold = np.nansum([position[f"STC{i}-uQty"] for i in range(1,4)])
            if position["uQty"] - qty_sold == 0:
                self.portfolio.loc[open_trade, "isOpen"] = 0
                print(Back.GREEN + "Already sold")
                self.queue_prints.put(["Already sold", "", "green"])

                log_alert['action'] = f"{STC}-DoneBefore"
                log_alert["portfolio_idx"] = open_trade
                self.alerts_log = pd.concat([self.alerts_log, pd.DataFrame.from_records(log_alert, index=[0])], ignore_index=True)
                self.save_logs(["alert"])
                return

            # Sell all and close waiting stc orders
            if order['xQty'] == 1:                
                # Stop updater to avoid overlapping
                self.update_paused = True
                # Sell all and close waiting stc orders
                self.close_open_exit_orders(open_trade)
                
                # if no uQty get all remaining
                if order['uQty'] is None:
                    position = self.portfolio.iloc[open_trade]
                    order['uQty'] = int(position["uQty"]) - qty_sold

            elif order['xQty'] < 1:  # portion
                # Stop updater to avoid overlapping
                self.update_paused = True
                self.close_open_exit_orders(open_trade)
                order['uQty'] = round(max(qty_bought * order['xQty'], 1))

            if order['uQty'] + qty_sold > qty_bought:
                order['uQty'] = qty_bought - qty_sold
                str_msg = f"Order {order['Symbol']} Qty exceeded, changed to {order['uQty']}"
                print(Back.RED + Fore.BLACK + str_msg)
                self.queue_prints.put([str_msg, "", "red"])

            order_response, order_id, order, _ = self.confirm_and_send(order, pars, self.bksession.make_STC_lim)
            log_alert["portfolio_idx"] = open_trade

            if order_response is None:  # Assume trade rejected by user
                log_alert['action'] = "STC-notAccepted"
                self.alerts_log = pd.concat([self.alerts_log, pd.DataFrame.from_records(log_alert, index=[0])], ignore_index=True)
                self.save_logs(["alert"])
                print(Back.GREEN + "STC not accepted by user, order response null")
                self.queue_prints.put(["STC not accepted by user, order response null", "", "green"])
                self.update_paused = False
                return

            order_status, order_info = self.get_order_info(order_id)
            self.portfolio.loc[open_trade, STC + "-ordID"] = order_id
            self.portfolio.loc[open_trade, STC + "-Price-Current"] = order["price_current"]

            # Check if STC price changed
            if order_status in ["FILLED", 'EXECUTED', 'INDIVIDUAL_FILLS']:
                self.log_filled_STC(order_id, open_trade, STC)
            else:
                str_STC = f"Submitted: {STC} {order['Symbol']} @{order['price']} Qty:{order['uQty']} ({order['xQty']})"
                print(Back.GREEN + str_STC)
                self.queue_prints.put([str_STC, "", "green"])

            #Log trades_log
            log_alert['action'] = "STC-partial" if order['xQty']<1 else "STC-ALL"
            self.alerts_log = pd.concat([self.alerts_log, pd.DataFrame.from_records(log_alert, index=[0])], ignore_index=True)
            self.save_logs()

            self.update_paused = False


    def log_filled_STC(self, order_id, open_trade, STC):

        order_status, order_info = self.get_order_info(order_id)

        sold_unts = order_info['orderLegCollection'][0]['quantity']

        if 'price' in order_info.keys():
            stc_price = order_info['price']
        elif 'stopPrice' in order_info.keys():
            stc_price = order_info['stopPrice']
        elif "orderActivityCollection" in order_info.keys():
            prics = []
            for ind in order_info["orderActivityCollection"]:
                prics.append([ind['quantity'], ind['executionLegs'][0]['price']])
                n_tot= sum([i[0] for i in prics])
            stc_price =  sum([i[0]*i[1] for i in prics])/ n_tot

        bto_price = self.portfolio.loc[open_trade, "Price"]
        bto_price_alert = self.portfolio.loc[open_trade, "Price-Alert"]
        bto_price_current = self.portfolio.loc[open_trade, "Price-Current"]
        stc_PnL = float((stc_price - bto_price)/bto_price) *100

        xQty = sold_unts/ self.portfolio.loc[open_trade, "uQty"]

        date = order_info["closeTime"]
        #Log portfolio
        self.portfolio.loc[open_trade, STC + "-Status"] = order_status
        self.portfolio.loc[open_trade, STC + "-Price"] = stc_price
        self.portfolio.loc[open_trade, STC + "-Date"] = date
        self.portfolio.loc[open_trade, STC + "-xQty"] = xQty
        self.portfolio.loc[open_trade, STC + "-uQty"] = sold_unts
        self.portfolio.loc[open_trade, STC + "-PnL"] = stc_PnL
        self.portfolio.loc[open_trade, STC + "-ordID"] = order_id

        trade = self.portfolio.loc[open_trade]
        sold_tot = np.nansum([trade[f"STC{i}-uQty"] for i in range(1,4)])
        stc_PnL_all = np.nansum([trade[f"STC{i}-PnL"]*trade[f"STC{i}-uQty"] for i in range(1,4)])/sold_tot
        self.portfolio.loc[open_trade, "PnL"] = stc_PnL_all
        stc_PnL_all_alert =  np.nansum([(float((trade[f"STC{i}-Price-Alerted"] - bto_price_alert)/bto_price_alert) *100) * trade[f"STC{i}-uQty"] for i in range(1,4)])/sold_tot
        stc_PnL_all_curr = np.nansum([(float((trade[f"STC{i}-Price-Current"] - bto_price_current)/bto_price_current) *100) * trade[f"STC{i}-uQty"] for i in range(1,4)])/sold_tot
        self.portfolio.loc[open_trade, "PnL-Alert"] = stc_PnL_all_alert
        self.portfolio.loc[open_trade, "PnL-Current"] = stc_PnL_all_curr
        
        mutipl = 1 if trade['Asset'] == "option" else .01  # pnl already in %
        self.portfolio.loc[open_trade, "$PnL"] =  stc_PnL_all* bto_price *mutipl*sold_tot
        self.portfolio.loc[open_trade, "$PnL-Alert"] =  stc_PnL_all_alert* bto_price_alert *mutipl*sold_tot
        self.portfolio.loc[open_trade, "$PnL-Current"] =  stc_PnL_all_curr* bto_price_current *mutipl*sold_tot
        
        
        symb = self.portfolio.loc[open_trade, 'Symbol']

        sold_Qty =  self.portfolio.loc[open_trade, [f"STC{i}-uQty" for i in range(1,4)]].sum()

        str_STC = f"{STC} {symb} @{stc_price} Qty:" + \
            f"{sold_unts}({int(xQty*100)}%), for {stc_PnL:.2f}%"

        if sold_Qty == self.portfolio.loc[open_trade, "uQty"]:
            str_STC += " (Closed)"
            self.portfolio.loc[open_trade, "isOpen"] = 0

        print (Back.GREEN + f"Filled: {str_STC}")
        self.queue_prints.put([f"Filled: {str_STC}","", "green"])
        self.save_logs()


    def update_orders(self):

        for i in range(len(self.portfolio)):
            self.close_expired(i)
            trade = self.portfolio.iloc[i]
            redo_orders = False

            if trade["isOpen"] == 0:
                continue

            if trade["BTO-Status"]  in ["QUEUED", "WORKING", 'OPEN']:
                _, order_info = self.get_order_info(trade['ordID'])

                # Check if number filled Qty changed
                qty_fill = order_info['filledQuantity']
                qty_fill_old = self.portfolio.loc[i, "filledQty"]
                # If so, redo orders
                if not (pd.isnull(qty_fill_old) or qty_fill_old == 0) and \
                    qty_fill_old != qty_fill:
                    redo_orders = True

                self.portfolio.loc[i, "filledQty"] = order_info['filledQuantity']
                self.portfolio.loc[i, "BTO-Status"] = order_info['status']
                trade = self.portfolio.iloc[i]
                self.save_logs("port")

            if pd.isnull(trade["filledQty"]) or trade["filledQty"] == 0:
                continue

            if trade.get("BTO-avg-Status") in ["QUEUED", "WORKING", 'OPEN']:
                ordID = trade['ordID'].split(",")[-1]
                _, order_info = self.get_order_info(ordID)
                if order_info['status'] == 'FILLED' :
                    self.portfolio.loc[i, "BTO-avg-Status"] = order_info['status']
                    self.portfolio.loc[i, "filledQty"] += order_info['filledQuantity']
                    redo_orders = True
                    trade = self.portfolio.iloc[i]
                    self.save_logs("port")

            if redo_orders:
                self.close_open_exit_orders(i)

            exit_plan = eval(trade["exit_plan"])
            if  exit_plan != {}:                
                if any([isinstance(e, str) and "%" not in e for e in exit_plan.values()]) and trade['Asset'] == 'option':
                    self.check_opt_stock_price(i, exit_plan, "STC")
                else:
                    self.make_exit_orders(i, exit_plan)

            # Go over STC orders and check status
            for ii in range(1, 4):
                STC = f"STC{ii}"
                trade = self.portfolio.iloc[i]
                STC_ordID = trade[STC+"-ordID"]

                if pd.isnull(STC_ordID):
                    continue

                # Get status exit orders
                STC_ordID = int(float(STC_ordID))  # Might be read as a float

                order_status, _ =  self.get_order_info(STC_ordID)

                if order_status == 'CANCELED':
                    # Try next order number. probably went through. This is for TDA OCO
                    order_status, _ =  self.get_order_info(STC_ordID + 1)
                    if order_status == 'FILLED':
                        STC_ordID = STC_ordID + 1
                        self.portfolio.loc[i, STC + "-ordID"] =  STC_ordID

                self.portfolio.loc[i, STC+"-Status"] = order_status
                trade = self.portfolio.iloc[i]

                if order_status in ["FILLED", "EXECUTED"] and np.isnan(trade[STC+"-xQty"]):
                    self.log_filled_STC(STC_ordID, i, STC)

        self.save_logs("port")


    def check_opt_stock_price(self, open_trade, exit_plan, act="STC"):
        "Option exits in stock price"
        i = open_trade
        exit_plan_ori = exit_plan.copy()
        trade = self.portfolio.iloc[i]

        ord_inf = trade['Symbol'].split("_")
        if len(ord_inf) != 2: return
        symb_stock = ord_inf[0]

        quote = self.price_now(symb_stock, act, 1)
        quote_opt = self.price_now(trade['Symbol'], act, 1)

        for v, pt in exit_plan.items():
            if not isinstance(pt, str): continue
            if v[:2] == "PT" and float(pt) <= quote:
                exit_plan[v] = quote_opt
                # Add another exit plan for x2
                STCn = int(v[2])
                if STCn < 3 and exit_plan[f"PT{STCn+1}"] is None:
                    exit_plan[f"PT{STCn+1}"] = quote_opt * 2
            elif v[:2] == "SL" and "%" not in pt and float(pt) >= quote:
                 exit_plan[v] = quote

        if exit_plan_ori != exit_plan:
            self.portfolio.loc[i, "exit_plan"] = str(exit_plan)
            self.save_logs("port")
            self.make_exit_orders(i, exit_plan)


    def SL_below_market(self, order, new_SL_ratio=.95):
        SL = order.get("SL")
        price_now = self.price_now(order["Symbol"], "STC", 1)

        if SL > price_now:
            new_SL = round(price_now * new_SL_ratio, 2)
            print(Back.RED + f"{order['Symbol']} SL below bid price, changed from {SL} to {new_SL}")
            self.queue_prints.put([f"{order['Symbol']} SL below bid price, changed from {SL} to {new_SL}",
                            "", "red"])
            order["SL"] = new_SL
        return order


    def make_exit_orders(self, open_trade, exit_plan):
        i = open_trade
        trade = self.portfolio.iloc[i]

        # Calculate x/uQty:
        uQty_bought = trade['filledQty']
        nPTs =  len([i for i in range(1,4) if exit_plan[f"PT{i}"] is not None])
        if nPTs != 0:
            uQty = [round(uQty_bought/nPTs)]*nPTs
            uQty[-1] = int(uQty_bought - sum(uQty[:-1]))

            xQty = [round(1/nPTs, 1)]*nPTs
            xQty[-1] = 1 - sum(xQty[:-1])

        # Go over exit plans and make orders
        order = {'Symbol': trade['Symbol']}
        for ii in range(1, nPTs+1):
            STC = f"STC{ii}"           

            # If Option add strike field
            ord_inf = trade['Symbol'].split("_")
            if len(ord_inf) == 2:
                opt_type = "C" if "C" in ord_inf[1] else "P"
                strike = str(re.split('C|P', ord_inf[1])[1]) + opt_type
                order['strike'] = strike

            STC_ordID = trade[STC+"-ordID"]
            if not pd.isnull(STC_ordID):
                # Adjust if necessary uQty based on remaining shares
                if nPTs < 2:
                    continue
                ord_stat, ord_inf = self.get_order_info(int(STC_ordID))
                iord_qty = ord_inf.get('quantity')
                if iord_qty is None:
                    iord_qty = ord_inf['childOrderStrategies'][0]['quantity']
                if uQty[ii - 1] != iord_qty:
                    uQty[ii - 1] = int(iord_qty)
                    uleft = uQty_bought - sum(uQty[:ii])
                    if uleft == 0:
                        break
                    nPts = len(uQty) - ii
                    if nPts !=0:
                        uQty = uQty[:ii] + [round(uleft/nPts)]*nPts
                        uQty[-1] = int(uQty_bought - sum(uQty[:-1]))
                    else:
                        uQty[-1] = int(uleft)
                    xQty = [round(u/uQty_bought,1) for u in uQty]

            else:
                SL = exit_plan["SL"]
                # Check if exit prices are strings (stock price for option)
                if isinstance(SL, str) and "%" not in SL: SL = None
                if isinstance(exit_plan[f"PT{ii}"], str): exit_plan[f"PT{ii}"] = None

                ord_func = None
                # Lim and Sl OCO order
                if exit_plan[f"PT{ii}"] is not None and SL is not None:
                    # Lim_SL order
                    ord_func = self.bksession.make_Lim_SL_order
                    order["PT"] = exit_plan[f"PT{ii}"]
                    order["SL"] = exit_plan["SL"]
                    order['uQty'] = uQty[ii - 1]
                    order['xQty'] = xQty[ii - 1]

                # Lim order
                elif exit_plan[f"PT{ii}"] is not None and SL is None:
                    ord_func = self.bksession.make_STC_lim
                    order["price"] = exit_plan[f"PT{ii}"]
                    order['uQty'] = uQty[ii - 1]
                    order['xQty'] = xQty[ii - 1]

                # SL order
                elif ii == 1 and SL is not None:
                    if "%" in SL:
                        ord_func = self.bksession.make_STC_SL_trailstop
                        order["trail_stop_percent"] = float(exit_plan["SL"].replace("%", ""))
                    else:
                        ord_func =self.bksession.make_STC_SL
                        order["price"] = exit_plan["SL"]
                    
                    order['uQty'] = int(trade['uQty'])
                    order['xQty'] = 1

                elif ii > 1 and SL is not None:
                    break

                else:
                    raise("Case not caught")

                # Check that is below current price
                if order.get("SL") is not None and "%" not in order.get("SL"):
                    order = self.SL_below_market(order)

                if ord_func is not None and order['uQty'] > 0:
                    _, STC_ordID = self.bksession.send_order(ord_func(**order))
                    if order.get("price"):
                        str_prt = f"{STC} {order['Symbol']} @{order['price']}(Qty:{order['uQty']}) sent during order update"
                    else:
                        str_prt = f"{STC} {order['Symbol']} @{order.get('PT')}/{order.get('SL')} (Qty:{order['uQty']}) sent during order update"
                    print (Back.GREEN + str_prt)
                    self.queue_prints.put([str_prt,"", "green"])
                    self.portfolio.loc[i, STC+"-ordID"] = STC_ordID
                    trade = self.portfolio.iloc[i]
                    self.save_logs("port")
                else:
                    break
        # no PTs but trailing stop
        if nPTs == 0 and exit_plan["SL"] is not None and "%" in exit_plan["SL"] and pd.isnull(trade["STC1-ordID"]):
            order["trail_stop_percent"] = float(exit_plan["SL"].replace("%", ""))      
            order['uQty'] = int(trade['uQty'])
            order['xQty'] = 1
            _, STC_ordID = self.bksession.send_order(self.bksession.make_STC_SL_trailstop(**order))
            str_prt = f"STC1 {order['Symbol']} Trailing stop of {order['trail_stop_percent']}% sent during order update"            
            print(Back.GREEN + str_prt)
            self.queue_prints.put([str_prt,"", "green"])
            self.portfolio.loc[i, "STC1-ordID"] = STC_ordID
            trade = self.portfolio.iloc[i]
            self.save_logs("port")


    def close_expired(self, open_trade):
        i = open_trade
        trade = self.portfolio.iloc[i]
        if trade["Asset"] != "option" or trade["isOpen"] == 0:
            return
        optdate = option_date(trade['Symbol'])
        if optdate.date() < date.today():
            expdate = date.today().strftime("%Y-%m-%dT%H:%M:%S+0000")
            usold = np.nansum([trade[f"STC{i}-uQty"] for i in range(1,4)])
            for stci in range(1,4):
                if pd.isnull(trade[f"STC{stci}-uQty"]):
                    STC = f"STC{stci}"
                    break

            #Log portfolio
            self.portfolio.loc[open_trade, STC + "-Status"] = 'EXPIRED'
            self.portfolio.loc[open_trade, STC + "-Price"] = 0
            self.portfolio.loc[open_trade, STC + "-Date"] = expdate
            self.portfolio.loc[open_trade, STC + "-xQty"] = 1
            self.portfolio.loc[open_trade, STC + "-uQty"] = trade['filledQty'] - usold
            self.portfolio.loc[open_trade, STC + "-PnL"] = -100
            self.portfolio.loc[open_trade, "isOpen"] = 0
            
            bto_price = self.portfolio.loc[open_trade, "Price"]
            bto_price_alert = self.portfolio.loc[open_trade, "Price-Alert"]
            bto_price_current = self.portfolio.loc[open_trade, "Price-Current"]
            
            trade = self.portfolio.loc[open_trade]
            sold_tot = np.nansum([trade[f"STC{i}-uQty"] for i in range(1,4)])
            stc_PnL_all = np.nansum([trade[f"STC{i}-PnL"]*trade[f"STC{i}-uQty"] for i in range(1,4)])/sold_tot
            self.portfolio.loc[open_trade, "PnL"] = stc_PnL_all
            stc_PnL_all_alert =  np.nansum([(float((trade[f"STC{i}-Price-Alerted"] - bto_price_alert)/bto_price_alert) *100) * trade[f"STC{i}-uQty"] for i in range(1,4)])/sold_tot
            stc_PnL_all_curr = np.nansum([(float((trade[f"STC{i}-Price-Current"] - bto_price_current)/bto_price_current) *100) * trade[f"STC{i}-uQty"] for i in range(1,4)])/sold_tot
            self.portfolio.loc[open_trade, "PnL-Alert"] = stc_PnL_all_alert
            self.portfolio.loc[open_trade, "PnL-Current"] = stc_PnL_all_curr
            
            mutipl = 1 if trade['Asset'] == "option" else .01  # pnl already in %
            self.portfolio.loc[open_trade, "$PnL"] =  stc_PnL_all* bto_price *mutipl*sold_tot
            self.portfolio.loc[open_trade, "$PnL-Alert"] =  stc_PnL_all_alert* bto_price_alert *mutipl*sold_tot
            self.portfolio.loc[open_trade, "$PnL-Current"] =  stc_PnL_all_curr* bto_price_current *mutipl*sold_tot
        
            str_prt = f"{trade['Symbol']} option expired -100% uQty: {trade['filledQty']}"
            print(Back.GREEN + str_prt)
            self.queue_prints.put([str_prt,"", "green"])
            self.save_logs("port")

def option_date(opt_symbol):
    sym_inf = opt_symbol.split("_")[1]
    opt_date = re.split("C|P", sym_inf)[0]
    return datetime.strptime(opt_date, "%m%d%y")

def amnt_left(order, position):
    # Calculate amnt to sell based on alerted left amount
    available = position["uQty"]
    if order.get("amnt_left"):
        left = order["amnt_left"]
        if left == "few":
            order['xQty'] =  1 - .2
            order['uQty'] = max(round(available * order['xQty']), 1)
        elif left > .99:  # unit left
            order['uQty'] = max(available - left, 1)
            order['xQty'] = (available - order['uQty'])/available
        elif left < .99:  # percentage left
            order['xQty'] = 1 - left
            order['uQty'] = max(round(available * order['xQty']), 1)
        else:
            raise ValueError
        return order, True
    else:
        return order, False

if __name__ == "__main__":
    from DiscordAlertsTrader.brokerages import get_brokerage
    
    bksession = get_brokerage()
    at = AlertsTrader(bksession)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Feb  8 18:11:55 2021

@author: adonay
"""
import re
import pandas as pd
from datetime import datetime
import numpy as np

def parse_trade_alert(msg, asset=None):
    pattern = r'\b(BTO|STC)\b\s*(\d+)?\s*([A-Z]+)\s*(\d+[.\d+]*[cp]?)?\s*(\d{1,2}\/\d{1,2})?(?:\/2023|\/23)?\s*@\s*[$]*[ ]*(\d+(?:[,.]\d+)?|\.\d+)'
    match = re.search(pattern, msg, re.IGNORECASE)
    
    if match:
        action, quantity, ticker, strike, expDate, price = match.groups()

        asset_type = 'option' if strike and expDate else 'stock'
        symbol =  ticker.upper()
        order = {
            'action': action.upper(),
            'Symbol': symbol,
            'uQty': int(quantity) if quantity else 1,
            'price': float(price.replace(',', '.')) if price else None,
            'asset': asset_type
        }
        str_ext = ""
        if asset_type == 'option':
            # fix missing strike, assume Call            
            if "c" not in strike.lower() and "p" not in strike.lower():
                strike = strike + "c"
                order['strike'] = strike.upper()
                str_ext = " No direction found in option strike, assuming Call"

            order['strike'] = strike.upper()
            order['expDate'] = expDate
            order['Symbol'] = fix_index_symbols(symbol)            
            order['Symbol'] = make_optionID(**order)

        risk_level = parse_risk(msg)
        order['risk'] = risk_level

        pars = []
        for el in [action, quantity, ticker, strike, expDate, price, risk_level, str_ext]:
            if el is not None:
                pars.append(el)
        pars = " ".join(pars)
        if action.upper() == "BTO":
            if "avg" in msg.lower() or "average" in msg.lower():
                avg_price, _ = parse_avg(msg)
                pars = pars + f"AVG to {avg_price} "
                order["avg"] = avg_price
            else:
                order["avg"] = None
            pt1_v, pt2_v, pt3_v, sl_v = parse_exits(msg)
            n_pts = 3 if pt3_v else 2 if pt2_v else 1 if pt1_v else 0
            pts_qty = set_pt_qts(n_pts)
            order, pars = make_order_exits(order, msg, pars, asset_type)
            sl_mental = True if "mental" in msg.lower() else False
            if sl_mental: order["SL_mental"] = True
            order["n_PTs"] = n_pts
            order["PTs_Qty"] = pts_qty

        elif action.upper() == "STC":
            xamnt = parse_sell_ratio_amount(msg, asset_type)
            if order["uQty"] is None:
                pars = pars + f" xamount: {xamnt}"
            order["xQty"] = xamnt
        
        return pars, order
    else:
        return None, None

def fix_index_symbols(symbol):
    if symbol.upper() == "SPX": 
        symbol = "SPXW"
    elif symbol.upper() == "NDX":
        symbol = "NDXP"
    return symbol
    

def parser_alerts(msg, asset=None):
    msg = msg.replace("STc", "STC").replace("StC", "STC").replace("stC", "STC").replace("STc", "STC")
    msg = msg.replace("BtO", "BTO").replace("btO", "BTO").replace("bTO", "BTO").replace("BTo", "BTO")
    msg = msg.replace("spx", "SPX").replace("spy", "SPY").replace("Spx", "SPX")
    act = parse_action(msg)
    if act is None:
        if "ExitUpdate" in msg:
            order = {}
            order['action'] = "ExitUpdate"
            order['Symbol'], _ = parse_Symbol(msg, "ExitUpdate")
            pars = f"ExitUpdate: "
            if asset == "option":
                if "_" not in  order['Symbol']:
                    strike, optType = parse_strike(msg)
                    order["expDate"] = parse_date(msg)
                    order["strike"] = strike + optType
                    order['Symbol'] = make_optionID(**order)
            order, str_prt = make_order_exits(order, msg, pars, "stock") # price type dealt in disc_trader
            return pars, order
        return None, None
    
    Symbol, Symbol_info = parse_Symbol(msg, act)
    if Symbol is None:
        return None, None

    if asset != "stock":
        strike, optType = parse_strike(msg)
        if asset is None:
            asset="option" if strike is not None else "stock"
        elif asset == "option" and strike is None:
            return None, None

    if asset == "option":
        expDate = parse_date(msg)
        if expDate is None:
            return None, None

        mark = parse_mark_option(msg)

    elif asset == "stock":
        mark  = parse_mark_stock(msg, Symbol, act)

    amnt = parse_unit_amount(msg)
    risk_level = parse_risk(msg)

    order = {"action": act,
             "Symbol": Symbol,
             "price": mark,
             "asset": asset,
             "uQty": amnt,
             "risk": risk_level}

    str_prt = f"{act} {Symbol} @{mark} amount: {amnt}"

    if asset == "option":
        if Symbol == "SPX": 
            order['Symbol'] = "SPXW"
        elif Symbol == "NDX":
            order['Symbol'] = "NDXP"
        optType = optType.upper()
        order["expDate"] = expDate
        order["strike"] = strike + optType
        str_prt = f"{act} {order['Symbol']} {expDate} {strike + optType} @{mark} amount: {amnt}"
        order['Symbol'] = make_optionID(**order)

    if act == "BTO":
        if "avg" in msg.lower() or "average" in msg.lower():
            avg_price, _ = parse_avg(msg)
            str_prt = str_prt + f"AVG to {avg_price} "
            order["avg"] = avg_price
        else:
            order["avg"] = None

        pt1_v, pt2_v, pt3_v, sl_v = parse_exits(msg)
        n_pts = 3 if pt3_v else 2 if pt2_v else 1 if pt1_v else 0
        pts_qty = set_pt_qts(n_pts)
        order, str_prt = make_order_exits(order, msg, str_prt, asset)
        sl_mental = True if "mental" in msg.lower() else False
        if sl_mental: order["SL_mental"] = True
        order["n_PTs"] = n_pts
        order["PTs_Qty"] = pts_qty

    elif act == "STC":
        xamnt = parse_sell_ratio_amount(msg, asset)
        if order["uQty"] is None:
            str_prt = str_prt + f" xamount: {xamnt}"
        order["xQty"] = xamnt
    return str_prt, order

def make_order_exits(order, msg, str_prt, asset):
    pt1_v, pt2_v, pt3_v, sl_v = parse_exits(msg)
    if asset == "option":
        order["PT1"] =  set_exit_price_type(pt1_v, order)
        order["PT2"] = set_exit_price_type(pt2_v, order)
        order["PT3"] = set_exit_price_type(pt3_v, order)
        order["SL"] = set_exit_price_type(sl_v, order)
    elif asset == "stock":
        order["PT1"] = pt1_v
        order["PT2"] = pt2_v
        order["PT3"] = pt3_v
        order["SL"] = sl_v

    exits = ["PT1","PT2","PT3","SL"]
    for ext in exits:
            if order["PT1"] is not None:
                str_prt = str_prt + f', {ext}:{order[ext]}'
    return order, str_prt

def set_exit_price_type(exit_price, order):
    """Option or stock price decided with smallest distance"""
    if exit_price is None:
        return exit_price
    if isinstance(exit_price, str): exit_price = eval(exit_price)

    price_strk = float(order['strike'][:-1])
    order_price = order.get('price', exit_price )  # IF NO ORDER PRICE TAKE EXIT!
    rtio_stock = abs(price_strk - exit_price)
    rtio_option = abs(order_price - exit_price)

    if rtio_stock < rtio_option:
        exit_price = str(exit_price)
    elif rtio_stock > rtio_option:
        exit_price = exit_price
    else:
        raise("Not sure if price of option or stock")
    return exit_price

def set_pt_qts(n_pts):
    if n_pts == 3:
        amnts = [.33, .33, .34]
    elif  n_pts == 2:
        amnts = [.5, .5]
    elif  n_pts in range(2):
        amnts = [1]
    return amnts

def parse_action(msg):
    if pd.isnull(msg):
        return None
    actions = ["BTO", "STC"]
    act =  actions[0] if actions[0] in msg  else actions[1] if actions[1] in msg else None
    return act

def parse_Symbol(msg, act):
    re_Symbol = re.compile("\*\*([A-Z]*?)\*\*")
    Symbol_info = re_Symbol.search(msg)

    if Symbol_info is None:
        re_Symbol = re.compile(f"{act} ([A-Z]+)")
        Symbol_info = re_Symbol.search(msg)

        if Symbol_info is None:
            for wrd in [ "ATH", "BTO", "STC", "ITM"]:
                msg = msg.replace(wrd+" ", " ")
            msg = msg.replace('VWAP', " ")
            msg = msg.replace("I'", "i'")

            re_Symbol = re.compile("([A-Z]+)(?![a-z])")
            Symbol_info = re_Symbol.search(msg)
            if Symbol_info is None:
                return None, None
            Symbol = Symbol_info.groups()[-1].replace(" ",'')

    Symbol = Symbol_info.groups()[-1]
    # print ("Symbol: ", Symbol)
    return Symbol, Symbol_info.span()

def parse_mark_stock(msg, Symbol, act):
    re_mark = re.compile("\@[ ]*[$]*[ ]*(\d+(?:\.\d+)?|\.\d+)")
    mark_inf = re_mark.search(msg)
    if mark_inf is None:
        re_mark = re.compile(f"{act} "+ "([\*])?([\*])?" + f"{Symbol}" +"([\*])?([\*])? (\d+(?:\.\d+)?)")
        mark_inf = re_mark.search(msg)
        if mark_inf is None:
            return None
    mark = float(mark_inf.groups()[-1])
    return mark

def parse_mark_option(msg):
    re_mark = re.compile("(?:@|at)[a-zA-Z]?[ ]*[$]?[ ]*([.]?\d+(?:\.\d+)?)")
    mark_inf = re_mark.search(msg)
    if mark_inf is None:
        re_mark = re.compile("(?:@|at)[a-zA-Z]?[ ]*[$]?[ ]*([,]?\d+(?:\.\d+)?)")
        mark_inf = re_mark.search(msg)
        
    if mark_inf is None:
        date = parse_date(msg)
        re_mark = re.compile(f"{date}[ ]*[$]?[ ]*([.]?\d+(?:\.\d+)?)")
        mark_inf = re_mark.search(msg)
        if mark_inf is None:
            return None
    if mark_inf.groups()[-1].count(".") > 1:
        if mark_inf.groups()[-1].startswith("."):
            return float(mark_inf.groups()[-1][1:].replace(",","."))
    mark = float(mark_inf.groups()[-1].replace(",","."))
    return mark

def parse_strike(msg):
    re_strike = re.compile(" [$]?(\d+(?:\.\d+)?)(C|c|P|p)")
    strike_inf = re_strike.search(msg)
    # avoid "BTO 1 COIN 73c" detected 1C
    if strike_inf is None:
        re_strike = re.compile(" [$]?(\d+(?:\.\d+)?)[ ]?(C|c|P|p)")
        strike_inf = re_strike.search(msg)
    elif strike_inf is None and "BTO" in msg:
        sym = parse_Symbol(msg, "BTO")[0]
        re_strike = re.compile(f"{sym} (\d+(?:\.\d+)?)")
        strike_inf = re_strike.search(msg)
        if strike_inf is None:
            return None, None
        return strike_inf.groups()[0], "C"

    if strike_inf is None:
        return None, None
    strike = strike_inf.groups()[0]
    optType = strike_inf.groups()[1].capitalize()
    return strike, optType

def parse_date(msg):
    # deal with 4 digit year
    re_date = re.compile("((\d{1,2}\/\d{1,2})\/(20\d{2}))")
    date_inf = re_date.search(msg)
    if date_inf is not None:
            dt_1 = date_inf.groups()[1]
            dt_2 = date_inf.groups()[2]
            date = f"{dt_1}/{dt_2[2:]}"
            return date
        
    re_date = re.compile("(\d{1,2}\/\d{1,2}(?:\/\d{1,2})?)")
    date_inf = re_date.search(msg)
    if date_inf is None:
        months = ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug",
                  "Sep", "Oct", "Nov", "Dec"]

        exp= f"({'|'.join(months)})" + " (\d{1,2}) (20\d{2})"
        re_date = re.compile(exp)
        date_inf = re_date.search(msg)

        if date_inf is None:
            # crx
            return None
        else:
            dt_1 = months.index(date_inf.groups()[0])
            dt_2 = date_inf.groups()[1]
            dt_3 = date_inf.groups()[2]
            date = f"{dt_1}/{dt_2}/{dt_3}"
            return date

    date = date_inf.groups()[0]
    return date

def parse_exits(msg):
    pt1_v = parse_exits_vals(msg, "PT[1]?")
    pt2_v = parse_exits_vals(msg, "PT2")
    pt3_v = parse_exits_vals(msg, "PT3")
    sl_v = parse_exits_vals(msg, "SL(?: below)?")
    return pt1_v, pt2_v, pt3_v, sl_v

def parse_avg(msg):
    re_avg = re.compile("(?:avg[.]?|new average)[ ]*[$]*(\d+(?:\.\d+)?)", re.IGNORECASE)
    avg_inf = re_avg.search(msg.lower())
    if avg_inf is None:
        return None, None
    avg = float(avg_inf.groups()[-1])
    return avg, avg_inf.span()

def parse_exits_vals(msg, expr):
    re_comp= re.compile("(" + expr + "[:]?[ ]*[$]*(\d+[\.]*[\d]*))", re.IGNORECASE)
    exit_inf = re_comp.search(msg)

    if exit_inf is None:
        re_comp= re.compile("(" + expr.lower() + "[:]?[ ]*[$]*(\d+[\.]*[\d]*))")
        exit_inf = re_comp.search(msg)

        if exit_inf is None:
            return None

    exit_v = float(exit_inf.groups()[-1].replace("..", ""))
    return exit_v

def parse_unit_amount(msg):    
    act = parse_action(msg)
    Symbol, _ = parse_Symbol(msg, act)
    
    exprs = f"{act}\s+(\d+) {Symbol}"
    re_comp= re.compile(exprs, re.IGNORECASE)
    amnt_inf = re_comp.search(msg)
    if amnt_inf is not None:
        return round(eval(amnt_inf.groups()[0]), 2)
    return

def parse_sell_ratio_amount(msg, asset):    
    exprs = "(?:sold|sell) (\d\/\d)"
    re_comp= re.compile(exprs, re.IGNORECASE)
    amnt_inf = re_comp.search(msg)
    if amnt_inf is not None:
        return round(eval(amnt_inf.groups()[0]), 2)

    exprs = "(?:sold|sell)(\d of \d)"
    re_comp= re.compile(exprs, re.IGNORECASE)
    amnt_inf = re_comp.search(msg)
    if amnt_inf is not None:
        return round(eval(amnt_inf.groups()[0].replace(" of ", "/")), 2)

    exprs = "(?:sold|sell) (\d{1,2})%"
    re_comp= re.compile(exprs, re.IGNORECASE)
    amnt_inf = re_comp.search(msg)
    if amnt_inf is not None:
        return round(float(amnt_inf.groups()[0])/100, 2)

    if any(subs in msg.lower() for subs in ["half off my remaining position", "selling half off"]):
        return 0.25

    if any(subs in msg.lower() for subs in ["sold half", "sold another half", "half"]):
        return 0.5

    exprs = "\((\d(?:\/| of )\d)\)"
    re_comp= re.compile(exprs)
    amnt_inf = re_comp.search(msg)
    if amnt_inf is not None:
        return round(eval(amnt_inf.groups()[0].replace(" of ", "/")), 2)

    partial = ['scaling out', 'selling more', 'trimming more off', "selling some more"]
    if any([True if m in msg.lower() else False for m in partial]):
        return .33

    if "partial" in msg.lower():
        if asset == "stock":
            amnt = .33
        elif asset == "option":
            amnt = .5
    else:
        amnt = 1
    return amnt

def parse_risk(msg):
    risk = {'very high risk':"very high",
            'risk very high':"very high",
            'very risky':"very high",
            'risk high': "high",
            'high risk': "high",
            'lotto': "lotto",
            'risky': "medium",
            'yolo':"yolo"}
    risk_level = None
    if "BTO" in msg:
        for k, rsk in risk.items():
            if k in msg.lower():
                risk_level = rsk
                break
    return risk_level

def parse_exit_plan(order):
    exit_plan = {}
    for p in [f"PT{i}" for i in range (1,4)] + ["SL"]:
            exit_plan[p] = order.get(p)
    return exit_plan

def auhtor_parser(msg, author, asset):
    if author not in ['ScaredShirtless#0001', 'Kevin (Momentum)#4441']:
        new_order = {}
        def stc_amount(msg):
            ######### Leave N untis
            units = ["one", "two", "three"]
            units_left = f'(?:(?:L|l)eaving|leave)[ ]*(?:only|just)?[ ]*({"|".join(units)})'
            mtch = re.compile(units_left, re.IGNORECASE)
            mtch = mtch.search(msg)
            if mtch is not None:
                strnum = mtch.groups()[0]
                qty_left, = [i for i, v in enumerate(units) if v==strnum]
                return qty_left

            ######### Leave a few untis
            left_few = '(?:(?:L|l)eaving|leave)[ ]*(?:only|just)?[ ]*[a]? (few)'
            mtch = re.compile(left_few, re.IGNORECASE)
            mtch = mtch.search(msg)
            if mtch is not None:
                return "few"

            ######### Leave % amount
            left_perc = '(?:(?:L|l)eaving|leave) (?:about|only)?[ ]*(\d{1,2})%'
            mtch = re.compile(left_perc, re.IGNORECASE)
            mtch = mtch.search(msg)
            if mtch is not None:
                perc = mtch.groups()[0]
                return eval(perc)/100

            return None

        def match_exp(exp, msg):
            mtch = re.compile(exp, re.IGNORECASE)
            mtch = mtch.search(msg)
            if mtch is not None:
                return mtch.groups()[0]
            return None

        # in STC target might not be PT2
        if "STC" not in msg:
            pt1_exps = ['Target: (\d+[\.]*[\d]*)', 'target: (\d+[\.]*[\d]*)', 'target[a-zA-Z\s\,\.]*(\d+[\.]*[\d]*)',
                   '(\d+[\.]*[\d]*)[a-zA-Z\s\,\.]*target',
                   'looking for (\d+[\.]*[\d]*)']
            for exp in pt1_exps:
                pt1 = match_exp(exp, msg)
                if pt1:
                    pt1 = pt1[:-1] if pt1[-1] == '.' else pt1
                    new_order["PT1"] = pt1
                    msg = msg.replace(pt1, " ")
                    break

            pt2 = "target[a-zA-Z0-9\.\s\,]*then (\d+[\.]*[\d]*)"
            pt2 = match_exp(pt2, msg)
            if pt2:
                pt2 = pt2[:-1] if pt2[-1] == '.' else pt2
                new_order["PT2"] = pt2
                msg = msg.replace(pt2, " ")
        else:
            pt2 = "second target[a-zA-Z0-9\.\s\,]*(\d+[\.]*[\d]*)"
            pt2 = match_exp(pt2, msg)
            if pt2:
                new_order["PT2"] = pt2
                msg = msg.replace(pt2, " ")

        sl_exps = ['Stop: (\d+[\.]*[\d]*)', 'stop: (\d+[\.]*[\d]*)', '(\d+[\.]*[\d]*)[a-zA-Z\s\,\.]{0,5}?stop',
                   'stop[a-zA-Z\s\,\.]*(\d+[\.]*[\d]*)']
        for exp in sl_exps:
            sl = match_exp(exp, msg)
            if sl:
                sl = sl[:-1] if sl[-1] == '.' else sl
                new_order["SL"] = sl
                break

        if "BTO" not in msg:
            amnt_left = stc_amount(msg)
            if amnt_left:
                new_order["amnt_left"] = amnt_left

            if "STC" not in msg:
                stc = "([^a-z]selling|[^a-z]sold|all out|(:?(out|took)[a-zA-Z\s]*last)|sell here|took some off)"
                mtch = re.compile(stc, re.IGNORECASE)
                mtch = mtch.search(msg)
                no_Sell = ["not selling yet", "Over Sold", "contracts sold", "How many sold it?", 'No need to ask me if I’m selling']
                if mtch is not None and not any([True if s in msg else False for s  in no_Sell]):
                    new_order['action'] = "STC"
                    new_order["xQty"]  = parse_sell_ratio_amount(msg, asset)

        if len(list(new_order.values())):
            symbol, _ = parse_Symbol(msg, parse_action(msg))
            if symbol:
                new_order["Symbol"] = symbol
            return new_order
        else:
            return None
    return None

def get_symb_prev_msg(df_hist, msg_ix, author):
    # df_hist["Author"]  = df_hist["Author"].apply(lambda x: x.split("#")[0])

    df_hist_auth = df_hist[df_hist["Author"]==author]
    msg_inx_auth, = np.nonzero(df_hist_auth.index == msg_ix)
    indexes = df_hist_auth.index.values

    for n in range(1,6):
        inx = indexes[msg_inx_auth - n]
        msg, = df_hist_auth.loc[inx, 'Content'].values
        if pd.isnull(msg):
            continue
        symbol, _ = parse_Symbol(msg, parse_action(msg))
        if symbol is not None:
            return symbol, inx
    return None, None

def make_optionID(Symbol:str, expDate:str, strike=str, **kwarg):
    """
    date: "[M]M/[D]D" or "[M]M/[D]D/YY[YY]"
    """
    strike, opt_type = float(strike[:-1]), strike[-1]
    date_elms = expDate.split("/")
    date_frm = f"{int(date_elms[0]):02d}{int(date_elms[1]):02d}"
    if len(date_elms) == 2: # MM/DD, year = current year
        year = str(datetime.today().year)[-2:]
        date_frm = date_frm + year
    elif len(date_elms) == 3:
        date_frm = date_frm + f"{int(date_elms[2][-2:]):02d}"

    # Strike in interger if no decimals
    if strike == int(strike):
        return f"{Symbol}_{date_frm}{opt_type}{int(strike)}"
    return f"{Symbol}_{date_frm}{opt_type}{strike}"

def combine_new_old_orders(msg, order_old, pars, author, asset="option"):
    order_author = auhtor_parser(msg, author, asset)
    if order_author is None:
        return order_old, pars

    if order_old is not None:
        for k in order_author.keys():
            # If
            if order_author[k] == order_old.get(k) and k != "Symbol" or \
                order_author[k] != order_old.get(k) and k == "Symbol":
                if k == "Symbol":
                    # in case of ticker vs option symbol ID
                    if order_author[k] == order_old[k][:len(order_author[k])]:
                        order_author[k] = order_old[k]
                        continue
                resp = input(f"Found diff vals for {k}: new= {order_author[k]}, old= {order_old[k]} " +
                             "[1- new, 2- old, 0- break and fix]")
                if resp == '2':
                    order_author[k] = order_old.get(k)
                elif resp == '0':
                    raise "error"
        order = {**order_old, **order_author}
    else:
        order = order_author

    if order.get("action") is None:
        order["asset"] = asset
        exits = ["PT1", "PT2", "PT3", "SL"]
        if any([order.get(k) for k in exits]):
            order['action'] = "ExitUpdate"
            pars = f"ExitUpdate: {pars}"
            for ex in exits:
                val = order.get(ex)
                if val is not None:
                    pars = pars + f" {ex}:{val},"
    return order, pars



import os
import time
import pandas as pd
from datetime import datetime, timezone
import threading
from colorama import Fore, init
import discord # this is discord.py-self package not discord

from .message_parser import parse_trade_alert
from .configurator import cfg
from .configurator import channel_ids
from .alerts_trader import AlertsTrader
from .alerts_tracker import AlertsTracker
try:
    from .fend_bot import bot_msgs
    with_fend = True
except ImportError:
    with_fend = False


init(autoreset=True)

class dummy_queue():
    def __init__(self, maxsize=10):
        self.maxsize = maxsize
        self.queue = []

    def put(self, item):
        if len(self.queue) >= self.maxsize:
            self.queue.pop(0)
        self.queue.append(item)


class DiscordBot(discord.Client):
    def __init__(self, 
                 queue_prints=dummy_queue(maxsize=10), 
                 live_quotes=True, 
                 brokerage=None,
                 tracker_portfolio_fname=cfg['portfolio_names']["tracker_portfolio_name"]):
        super().__init__()
        self.channel_IDS = channel_ids
        self.time_strf = "%Y-%m-%d %H:%M:%S.%f"
        self.queue_prints = queue_prints
        self.bksession = brokerage
        self.live_quotes = live_quotes
        if brokerage is not None:
            self.trader = AlertsTrader(queue_prints=self.queue_prints, brokerage=brokerage)       
        self.tracker = AlertsTracker(brokerage=brokerage, portfolio_fname=tracker_portfolio_fname)
        self.load_data()        

        if live_quotes and brokerage is not None:
            self.thread_liveq =  threading.Thread(target=self.track_live_quotes)
            self.thread_liveq.start()

    def close_bot(self):
        if self.bksession is not None:
            self.trader.update_portfolio = False
            self.live_quotes = False

    def track_live_quotes(self):
        dir_quotes = cfg['general']['data_dir'] + '/live_quotes'
        os.makedirs(dir_quotes, exist_ok=True)

        while self.live_quotes:
            # Skip closed market
            now = datetime.now()
            weekday, hour = now.weekday(), now.hour
            if  weekday >= 5 or (hour < 9 and hour >= 17):  
                time.sleep(60)
                continue

            # get unique symbols  from portfolios
            track_symb = set(self.tracker.portfolio.loc[self.tracker.portfolio['isOpen']==1, 'Symbol'].to_list() + \
                self.trader.portfolio.loc[self.trader.portfolio['isOpen']==1, 'Symbol'].to_list())
            # save quotes to file
            try:
                quote = self.bksession.get_quotes(track_symb)
            except Exception as e:
                print('error during live quote:', e)
                continue
            
            for q in quote: 
                if quote[q]['description'] == 'Symbol not found':
                    continue
                timestamp = quote[q]['quoteTimeInLong']//1000  # in ms
                do_header = not os.path.exists(f"{dir_quotes}/{quote[q]['symbol']}.csv")
                with open(f"{dir_quotes}/{quote[q]['symbol']}.csv", "a+") as f:
                    if do_header:
                        f.write(f"timestamp, quote\n")
                    f.write(f"{timestamp}, {quote[q]['bidPrice']}\n")
            
            # Sleep for up to 5 secs    
            toc = (datetime.now() - now).total_seconds()
            if toc < 5 and self.live_quotes:
                time.sleep(5-toc)

    def load_data(self):
        self.chn_hist= {}
        self.chn_hist_fname = {}
        for ch in self.channel_IDS.keys():
            dt_fname = f"{cfg['general']['data_dir']}/{ch}_message_history.csv"
            if not os.path.exists(dt_fname):
                ch_dt = pd.DataFrame(columns=cfg['col_names']['chan_hist'].split(","))
                ch_dt.to_csv(dt_fname, index=False)
                ch_dt.to_csv(f"{cfg['general']['data_dir']}/{ch}_message_history_temp.csv", index=False)
            else:
                ch_dt = pd.read_csv(dt_fname)

            self.chn_hist_fname[ch] = dt_fname
            self.chn_hist[ch]= ch_dt

    async def on_ready(self):
        print('Logged on as', self.user , '\n loading previous messages')
        await self.load_previous_msgs()

    async def on_message(self, message):
        # handle fend bot messages
        if with_fend:
            alert = bot_msgs(message)
            if alert is not None:
                self.new_msg_acts(alert, False)
                return
        # only respond to channels in config        
        if message.channel.id not in self.channel_IDS.values():
            return
        if message.content == 'ping':
            await message.channel.send('pong')
        if not len(message.content):
            return
        self.new_msg_acts(message)

    async def on_message_edit(self, before, after):
        # Ignore if the message is not from a user or if the bot itself edited the message
        if after.channel.id not in self.channel_IDS.values() or  before.author.bot:
            return

        str_prt = f"Message edited by {before.author}: '{before.content}' -> '{after.content}'"
        self.queue_prints.put([str_prt, "blue"])
        print(Fore.BLUE + str_prt)

    async def load_previous_msgs(self):
        await self.wait_until_ready()
        for ch, ch_id in self.channel_IDS.items():
            channel = self.get_channel(ch_id)
            if channel is None:
                print("channel not found:", ch)
                continue
            
            if len(self.chn_hist[ch]):
                msg_last = self.chn_hist[ch].iloc[-1]
                date_After = datetime.strptime(msg_last.Date, self.time_strf) 
                iterator = channel.history(after=date_After, oldest_first=True)
            else:
                # iterator = channel.history(oldest_first=True)
                continue
                
            print("In", channel)
            async for message in iterator:
                self.new_msg_acts(message)
        print("Done")
        self.tracker.close_expired()

    def new_msg_acts(self, message, from_disc=True):
        if from_disc:
            msg_date = message.created_at.replace(tzinfo=timezone.utc).astimezone(tz=None)
            msg_date_f = msg_date.strftime(self.time_strf)    
            if message.channel.id in self.channel_IDS.values():
                chn_ix = list(self.channel_IDS.values()).index(message.channel.id)
                chn = list(self.channel_IDS.keys())[chn_ix]
            else:
                chn = None
            msg = pd.Series({'AuthorID': message.author.id,
                            'Author': f"{message.author.name}#{message.author.discriminator}",
                            'Date': msg_date_f, 
                            'Content': message.content,
                            'Channel': chn
                            })
        else:
            msg = message
        chn = msg['Channel']
        shrt_date = datetime.strptime(msg["Date"], self.time_strf).strftime('%Y-%m-%d %H:%M:%S')
        self.queue_prints.put([f"{shrt_date} \t {msg['Author']}: {msg['Content']} ", "blue"])
        print(Fore.BLUE + f"{shrt_date} \t {msg['Author']}: {msg['Content']} ")

        pars, order =  parse_trade_alert(msg['Content'])
        if pars is None:
            if self.chn_hist.get(chn) is not None:
                msg['Parsed'] = ""
                self.chn_hist[chn] = pd.concat([self.chn_hist[chn], msg.to_frame().transpose()],axis=0, ignore_index=True)
                self.chn_hist[chn].to_csv(self.chn_hist_fname[chn], index=False)
            return
        else:
            if order['asset'] == "option":
                # get option date with year
                opt_dt = datetime.strptime(f"{order['expDate']} {datetime.now().year}" , "%m/%d %Y")
                today = datetime.now().date()
                past = opt_dt.date() < today
                if past:
                    str_msg = f"Option date in the past: {order['expDate']}"
                    self.queue_prints.put([f"\t \t {str_msg}", "green"])
                    print(Fore.GREEN + f"\t \t {str_msg}")
                    msg['Parsed'] = str_msg
                    if self.chn_hist.get(chn) is not None:
                        self.chn_hist[chn] = pd.concat([self.chn_hist[chn], msg.to_frame().transpose()],axis=0, ignore_index=True)
                        self.chn_hist[chn].to_csv(self.chn_hist_fname[chn], index=False)
                    return
                
            order['Trader'], order["Date"] = msg['Author'], msg["Date"]
            order_date = datetime.strptime(order["Date"], "%Y-%m-%d %H:%M:%S.%f")
            date_diff = datetime.now() - order_date
            print(f"time difference is {date_diff.total_seconds()}")

            live_alert = True if date_diff.seconds < 90 else False
            str_msg = pars
            if live_alert and self.bksession is not None: 
                str_msg += " " + self.trader.price_now(order['Symbol'], order["action"], pflag=0)
            self.queue_prints.put([f"\t \t {str_msg}", "green"])
            print(Fore.GREEN + f"\t \t {str_msg}")
            
            track_out = self.tracker.trade_alert(order, live_alert, chn)
            self.queue_prints.put([f"\t \t tracker log: {track_out}", "red"])
            if msg['Author'] in cfg['discord']['authors_subscribed'] and  self.bksession is not None:
                order["Trader"] = msg['Author']
                if len(cfg["order_configs"]["default_trailstop"]):
                    order['SL'] = cfg["order_configs"]["default_trailstop"] + "%"
                self.trader.new_trade_alert(order, pars, msg['Content'])
        
        if self.chn_hist.get(chn) is not None:
            msg['Parsed'] = pars
            self.chn_hist[chn] = pd.concat([self.chn_hist[chn], msg.to_frame().transpose()],axis=0, ignore_index=True)
            self.chn_hist[chn].to_csv(self.chn_hist_fname[chn], index=False)



if __name__ == '__main__':
    client = DiscordBot()
    client.run(cfg['discord']['discord_token'])



